# -*- coding: utf-8 -*-

# Copyright the NTPsec project contributors
#
# SPDX-License-Identifier: BSD-2-Clause

import ntp.agentx as AX
import ntp.agentx_packet as AP
import ntp.util
import jigs

import unittest
import types


class TestMIBControl(unittest.TestCase):
    def test___init__(self):
        c = AX.MIBControl(1, 2, 3, 4, 5)
        self.assertEqual(c.oidTree, 1)
        self.assertEqual(c.inSetP, False)
        self.assertEqual(c.setVarbinds, [])
        self.assertEqual(c.setHandlers, [])
        self.assertEqual(c.setUndoData, [])
        self.assertEqual(c.mibRoot, 2)
        self.assertEqual(c.rangeSubid, 3)
        self.assertEqual(c.upperBound, 4)
        self.assertEqual(c.context, 5)

    def test_mib_rootOID(self):
        c = AX.MIBControl(mibRoot="foo")
        self.assertEqual(c.mib_rootOID(), "foo")

    def test_mib_rangeSubid(self):
        c = AX.MIBControl(rangeSubid="foo")
        self.assertEqual(c.mib_rangeSubid(), "foo")

    def test_mib_upperBound(self):
        c = AX.MIBControl(upperBound="foo")
        self.assertEqual(c.mib_upperBound(), "foo")

    def test_mib_context(self):
        c = AX.MIBControl(mibContext="foo")
        self.assertEqual(c.mib_context(), "foo")

    def test_addNode(self):
        c = AX.MIBControl()
        # Top level, non-dynamic
        c.addNode((0,), 1, 2)
        self.assertEqual(c.oidTree,
                         {0: {"reader": 1, "writer": 2, "subids": None}})
        # Top level, non-dynamic, class OID
        c.addNode(AP.OID((1,)), 1, 2)
        self.assertEqual(c.oidTree,
                         {0: {"reader": 1, "writer": 2, "subids": None},
                          1: {"reader": 1, "writer": 2, "subids": None}})
        # Top level, dynamic
        c.addNode((2,), dynamic=3)
        self.assertEqual(c.oidTree,
                         {0: {"reader": 1, "writer": 2, "subids": None},
                          1: {"reader": 1, "writer": 2, "subids": None},
                          2: {"reader": None, "writer": None, "subids": 3}})
        # Sub level
        c.addNode((1, 2, 3), 1, 2)
        self.assertEqual(c.oidTree,
                         {0: {"reader": 1, "writer": 2, "subids": None},
                          1: {"reader": 1, "writer": 2, "subids":
                              {2: {"reader": None, "writer": None, "subids": {
                                   3: {"reader": 1, "writer": 2, "subids": None}}}}},
                          2: {"reader": None, "writer": None, "subids": 3}})

    def test_getOID_andNext(self):
        c = AX.MIBControl()
        c.addNode((0, 1))
        c.addNode((0, 2, 0), 10, 11)
        c.addNode((0, 2, 1), 20, 21)
        c.addNode((0, 2, 2), 30, 31)
        c.addNode((0, 2, 3), 40, 41)
        c.addNode((0, 4, 1, 0), 50, 51)
        # Test getOID
        # Test non-existent, no generator
        self.assertEqual(c.getOID(AP.OID((0, 0))), (None, None, None))
        # Test empty, no generator
        self.assertEqual(c.getOID(AP.OID((0, 1))),
                         (None, None, None))
        # Test empty, with generator
        self.assertEqual(c.getOID(AP.OID((0, 1)), True),
                         (None, None, None, None))
        # Test existent, no generator
        self.assertEqual(c.getOID(AP.OID((0, 2, 1))),
                         (AP.OID((0, 2, 1)), 20, 21))
        # Test existent, with generator
        r = c.getOID(AP.OID((0, 2, 1)), True)
        self.assertEqual(r[:3], (AP.OID((0, 2, 1)), 20, 21))
        self.assertEqual(isinstance(r[3], types.GeneratorType), True)
        # Test getNextOID
        # Test non-existent, no generator
        self.assertEqual(c.getNextOID(AP.OID((0, 0))),
                         (AP.OID((0, 2, 0)), 10, 11))
        # Test empty, no generator
        self.assertEqual(c.getNextOID(AP.OID((0, 1))),
                         (AP.OID((0, 2, 0)), 10, 11))
        # Test empty, with generator
        r = c.getNextOID(AP.OID((0, 1)), True)
        self.assertEqual(r[:3], (AP.OID((0, 2, 0)), 10, 11))
        self.assertEqual(isinstance(r[3], types.GeneratorType), True)
        # Test existent, no generator
        self.assertEqual(c.getNextOID(AP.OID((0, 2, 1))),
                         (AP.OID((0, 2, 2)), 30, 31))
        # Test existent, with generator
        r = c.getNextOID(AP.OID((0, 2, 1)), True)
        self.assertEqual(r[:3], (AP.OID((0, 2, 2)), 30, 31))
        self.assertEqual(isinstance(r[3], types.GeneratorType), True)
        # Test walking off the end
        self.assertEqual(c.getNextOID(AP.OID((0, 4, 1, 0))),
                         (None, None, None))

    def test_getOIDsInRange(self):
        c = AX.MIBControl()
        c.addNode((0, 1))
        c.addNode((0, 2, 0), 10, 11)
        c.addNode((0, 2, 1), 20, 21)
        c.addNode((0, 2, 2), 30, 31)
        c.addNode((0, 2, 3), 40, 41)
        c.addNode((0, 4, 1, 0), 50, 51)
        # Test range too early
        rng = AP.SearchRange((0, 0, 0), (0, 0, 5))
        self.assertEqual(c.getOIDsInRange(rng), [])
        # Test range too late
        rng = AP.SearchRange((6, 0, 0), (6, 0, 5))
        self.assertEqual(c.getOIDsInRange(rng), [])
        # Test nothing implemented in range
        rng = AP.SearchRange((0,), (0, 1, 5))
        self.assertEqual(c.getOIDsInRange(rng), [])
        # Test in range
        rng = AP.SearchRange((0, 2, 0), (0, 2, 3))
        self.assertEqual(c.getOIDsInRange(rng),
                         [(AP.OID((0, 2, 1)), 20, 21),
                          (AP.OID((0, 2, 2)), 30, 31)])
        # Test unbounded range
        rng = AP.SearchRange((0, 2, 0), ())
        self.assertEqual(c.getOIDsInRange(rng),
                         [(AP.OID((0, 2, 1)), 20, 21),
                          (AP.OID((0, 2, 2)), 30, 31),
                          (AP.OID((0, 2, 3)), 40, 41),
                          (AP.OID((0, 4, 1, 0)), 50, 51)])


class TestPacketControl(unittest.TestCase):
    def test___init__(self):
        p = AX.PacketControl("sock", "base")
        self.assertEqual(p.socket, "sock")
        self.assertEqual(0.0005 < p.spinGap < 0.0015, True)
        self.assertEqual(p.packetLog, {})
        self.assertEqual(p.loopCallback, None)
        self.assertEqual(p.database, "base")
        self.assertEqual(p.receivedData, b"")
        self.assertEqual(p.timeout, 30)
        self.assertEqual(p.sessionID, None)
        self.assertEqual(p.highestTransactionID, 0)
        self.assertEqual(p.lastReception, None)
        self.assertEqual(p.stillConnected, False)
        self.assertEqual(p.pduHandlers,
                         {AP.PDU_GET: p.handle_GetPDU,
                          AP.PDU_GET_NEXT: p.handle_GetNextPDU,
                          AP.PDU_GET_BULK: p.handle_GetBulkPDU,
                          AP.PDU_TEST_SET: p.handle_TestSetPDU,
                          AP.PDU_COMMIT_SET: p.handle_CommitSetPDU,
                          AP.PDU_UNDO_SET: p.handle_UndoSetPDU,
                          AP.PDU_CLEANUP_SET: p.handle_CleanupSetPDU,
                          AP.PDU_RESPONSE: p.handle_ResponsePDU})

    def test_mainloop(self):
        stillCons = []
        loop_calls = [0]
        cb_calls = [0]

        def loop_jig(self):
            loop_calls[0] += 1
            self.stillConnected = stillCons.pop(0)

        def callback_jig(p):
            cb_calls[0] += 1

        p = AX.PacketControl(None, None)
        p.spinGap = 0
        p._doloop = (lambda: loop_jig(p))
        p.stillConnected = False
        # Test not connected
        self.assertEqual(p.mainloop(False), False)
        self.assertEqual(loop_calls, [0])
        # Test single shot
        p.stillConnected = True
        stillCons = [True] * 2
        self.assertEqual(p.mainloop(False), True)
        self.assertEqual(stillCons, [True])
        self.assertEqual(loop_calls, [1])
        # Test loop
        p.stillConnected = True
        loop_calls[0] = 0
        stillCons = [True] * 5 + [False]
        self.assertEqual(p.mainloop(True), False)
        self.assertEqual(stillCons, [])
        self.assertEqual(loop_calls, [6])
        # Test loop, with callback
        p.stillConnected = True
        loop_calls[0] = 0
        p.loopCallback = callback_jig
        stillCons = [True] * 5 + [False]
        self.assertEqual(p.mainloop(True), False)
        self.assertEqual(stillCons, [])
        self.assertEqual(loop_calls, [6])
        self.assertEqual(cb_calls, [6])

    def test__doloop(self):
        # need jigs for
        # time
        pkts = []
        # recpTimes = []
        handler_calls = [0]

        def packetEater_jig(self):
            self.receivedPackets = pkts
            self.lastReception = recpTime

        def pinghandler_jig(self):
            handler_calls[0] += 1

        faketimemod = jigs.TimeModuleJig()
        sock = jigs.SocketJig()
        p = AX.PacketControl(sock, None)
        p.packetEater = (lambda: packetEater_jig(p))
        p.checkResponses = (lambda: None)
        p.pduHandlers[AP.PDU_PING] = pinghandler_jig
        p.sessionID = 42
        try:
            timetemp = AX.time
            AX.time = faketimemod
            # Test first time, no reception
            recpTime = None
            faketimemod.time_returns = []
            p._doloop()
            self.assertEqual(sock.data, [])
            # Test wrong session ID
            pkts = [AP.PingPDU(True, 13, 0, 0)]
            recpTime = 50
            faketimemod.time_returns = [55]
            p._doloop()
            self.assertEqual(len(sock.data), 1)
            sentPkt = AP.decode_packet(sock.data[0])[0]
            respPkt = AP.ResponsePDU(True, 13, 0, 0, 0, AP.RSPERR_NOT_OPEN, 0)
            self.assertEqual(sentPkt, respPkt)
            # Test unimplemented packet type
            faketimemod.time_returns = [55]
            sock.data = []
            pkts = [AP.PingPDU(True, 42, 0, 0)]
            pkts[0].pduType = 1337  # spoof the type
            p._doloop()
            self.assertEqual(sock.data, [])
            self.assertEqual(handler_calls, [0])
            # Test successful
            faketimemod.time_returns = [55]
            sock.data = []
            pkts = [AP.PingPDU(True, 42, 0, 0)]
            p._doloop()
            self.assertEqual(sock.data, [])
            self.assertEqual(handler_calls, [1])
            # Test pinging
            handler_calls = [0]
            faketimemod.time_returns = [120]
            sock.data = []
            p._doloop()
            self.assertEqual(len(sock.data), 1)
            sentPkt = AP.decode_packet(sock.data[0])[0]
            respPkt = AP.PingPDU(True, 42, 5, 1)
            self.assertEqual(sentPkt, respPkt)
            self.assertEqual(handler_calls, [0])
        finally:
            AX.time = timetemp

    def test_initNewSession(self):
        resp_returns = []
        resp_calls = []

        def waitForResponse_jig(opkt, ignoreSID=False):
            resp_calls.append((opkt, ignoreSID))
            return resp_returns.pop(0)

        mjig = AX.MIBControl(mibRoot=(0, 1, 2), rangeSubid=0, upperBound=None,
                             mibContext="foo")
        sock = jigs.SocketJig()
        p = AX.PacketControl(sock, mjig)
        p.waitForResponse = waitForResponse_jig
        resp_returns = [AP.ResponsePDU(True, 42, 1, 1, 0, 0, 0),
                        AP.ResponsePDU(True, 42, 1, 1, 0, 0, 0)]
        p.initNewSession()
        self.assertEqual(resp_returns, [])
        self.assertEqual(len(sock.data), 2)
        openpkt = AP.decode_packet(sock.data[0])[0]
        registerpkt = AP.decode_packet(sock.data[1])[0]
        self.assertEqual(openpkt, AP.OpenPDU(True, 23, 0, 0, 30, (),
                                             "NTPsec SNMP subagent"))
        self.assertEqual(registerpkt, AP.RegisterPDU(True, 42, 1, 1, 30, 1,
                                                     mjig.mib_rootOID(),
                                                     mjig.mib_rangeSubid(),
                                                     mjig.mib_upperBound(),
                                                     mjig.mib_context()))
        self.assertEqual(p.sessionID, 42)
        self.assertEqual(p.stillConnected, True)

    def test_waitForResponse(self):
        pkts = [AP.PingPDU(True, 1, 2, 3),
                AP.ResponsePDU(True, 10, 20, 30, 0, 0, 0),
                AP.ResponsePDU(True, 5, 23, 100, 0, 0, 0),
                AP.ResponsePDU(True, 42, 23, 100, 0, 0, 0),
                AP.ResponsePDU(True, 100, 200, 300, 0, 0, 0)]
        opkt = AP.PingPDU(True, 42, 23, 100)

        def packetEater_jig(self):
            self.receivedPackets.append(pkts.pop(0))

        p = AX.PacketControl(None, None)
        p.packetEater = (lambda : packetEater_jig(p))
        p.spinGap = 0
        # Test with session id match required
        ret = p.waitForResponse(opkt)
        self.assertEqual(pkts, [AP.ResponsePDU(True, 100, 200, 300, 0, 0, 0)])
        self.assertEqual(ret, AP.ResponsePDU(True, 42, 23, 100, 0, 0, 0))
        # Test ignoring session id
        pkts = [AP.PingPDU(True, 1, 2, 3),
                AP.ResponsePDU(True, 10, 20, 30, 0, 0, 0),
                AP.ResponsePDU(True, 5, 23, 100, 0, 0, 0),
                AP.ResponsePDU(True, 42, 23, 100, 0, 0, 0)]
        ret = p.waitForResponse(opkt, True)
        self.assertEqual(pkts, [AP.ResponsePDU(True, 42, 23, 100, 0, 0, 0)])
        self.assertEqual(ret, AP.ResponsePDU(True, 5, 23, 100, 0, 0, 0))

    def test_checkResponses(self):
        returns = []
        callback = (lambda resp, pkt: returns.append((resp, pkt)))
        faketimemod = jigs.TimeModuleJig()
        p = AX.PacketControl(None, None)
        p.packetLog = {(42, 23, 100): (25, "foo", callback),
                       (42, 24, 200): (50, "bar", callback)}
        faketimemod.time_returns = [40]
        try:
            timetemp = AX.time
            AX.time = faketimemod
            p.checkResponses()
            self.assertEqual(returns, [(None, "foo")])
            self.assertEqual(p.packetLog,
                             {(42, 24, 200): (50, "bar", callback)})
        finally:
            AX.time = timetemp

    def test_packetEater(self):
        pkt = AP.ResponsePDU(True, 42, 23, 100, 0, 0, 0)
        enc = pkt.encode()
        part1, rest = ntp.util.slicedata(enc, 15)
        part2, part3 = ntp.util.slicedata(rest, 6)
        pollReturns = [part1, part2, part3 + b"blah"]

        def pollSocket_jig(self):
            self.receivedData += pollReturns.pop(0)
        p = AX.PacketControl(None, None)
        p.sessionID = 42
        p.pollSocket = (lambda : pollSocket_jig(p))
        # Test incomplete header
        p.packetEater()
        self.assertEqual(p.receivedPackets, [])
        self.assertEqual(p.receivedData, part1)
        # Test incomplete packet
        p.packetEater()
        self.assertEqual(p.receivedPackets, [])
        self.assertEqual(p.receivedData, part1 + part2)
        # Test packet eaten
        p.packetEater()
        self.assertEqual(p.receivedPackets, [pkt])
        self.assertEqual(p.receivedData, b"blah")

    def test_sendPacket(self):
        sock = jigs.SocketJig()
        p = AX.PacketControl(sock, None)
        testpkt = AP.PingPDU(True, 0, 1, 2, "fake")
        # Test not expecting reply
        p.sendPacket(testpkt, False)
        self.assertEqual(sock.data, [b"\x01\r\x18\x00"
                                     b"\x00\x00\x00\x00\x00\x00\x00\x01"
                                     b"\x00\x00\x00\x02\x00\x00\x00\x08"
                                     b"\x00\x00\x00\x04fake"])
        self.assertEqual(p.packetLog, {})
        sock.data = []
        # Test expecting reply
        p.sendPacket(testpkt, True, callback="foo")
        self.assertEqual(sock.data, [b"\x01\r\x18\x00"
                                     b"\x00\x00\x00\x00\x00\x00\x00\x01"
                                     b"\x00\x00\x00\x02\x00\x00\x00\x08"
                                     b"\x00\x00\x00\x04fake"])
        self.assertEqual(p.packetLog, {(0, 1, 2): (30.0, testpkt, "foo")})

    def test_sendPing(self):
        sock = jigs.SocketJig()
        p = AX.PacketControl(sock, None)
        p.sessionID = 42
        p.sendPing()
        self.assertEqual(sock.data, [b"\x01\r\x10\x00"
                                     b"\x00\x00\x00*\x00\x00\x00\x05"
                                     b"\x00\x00\x00\x01\x00\x00\x00\x00"])
        self.assertEqual(list(p.packetLog.keys()), [(42, 5, 1)])

    def test_sendNotify(self):
        sock = jigs.SocketJig()
        p = AX.PacketControl(sock, None)
        p.sessionID = 42
        vb = AP.Varbind(AP.VALUE_INTEGER, (0, 1, 2), 23)
        p.sendNotify([vb])
        self.assertEqual(sock.data,
                         [b"\x01\x0c\x10\x00"
                          b"\x00\x00\x00*\x00\x00\x00\x05"
                          b"\x00\x00\x00\x01\x00\x00\x00\x18"
                          b"\x00\x02\x00\x00"
                          b"\x03\x00\x00\x00\x00\x00\x00\x00"
                          b"\x00\x00\x00\x01\x00\x00\x00\x02"
                          b"\x00\x00\x00\x17"])
        self.assertEqual(list(p.packetLog.keys()), [(42, 5, 1)])

    def test_sendErrorResponse(self):
        sock = jigs.SocketJig()
        p = AX.PacketControl(sock, None)
        p.sessionID = 42
        header = {"flags": {"bigEndian": True}, "session_id": 42,
                  "transaction_id": 32, "packet_id": 100}
        p.sendErrorResponse(header, AP.ERR_GENERR, 12)
        self.assertEqual(sock.data,
                         [b"\x01\x12\x10\x00"
                          b"\x00\x00\x00*\x00\x00\x00 "
                          b"\x00\x00\x00d\x00\x00\x00\x08"
                          b"\x00\x00\x00\x00\x00\x05\x00\x0c"])
        self.assertEqual(p.packetLog, {})

    def test_pollSocket(self):
        fakeselectmod = jigs.SelectModuleJig()
        sock = jigs.SocketJig()
        p = AX.PacketControl(sock, None)
        p.sessionID = 42
        # Test
        pkt = AP.ResponsePDU(True, 42, 23, 100, 0, 0, 0)
        enc = pkt.encode()
        part1, part2 = ntp.util.slicedata(enc, 15)
        sock.return_data = [part1, part2, b"blah"]
        fakeselectmod.do_return = [True, True, True, False]
        try:
            selecttemp = AX.select
            AX.select = fakeselectmod
            p.pollSocket()
            self.assertEqual(p.receivedData, enc + b"blah")
        finally:
            AX.select = selecttemp

    def test_handleGetPDU(self):
        readReturns = [1]
        read = (lambda oid: AP.Varbind(AP.VALUE_INTEGER,
                                       oid, readReturns.pop(0)))
        mjig = AX.MIBControl()
        mjig.addNode((0, 0), read)
        mjig.addNode((0, 1), (lambda x: None))
        sock = jigs.SocketJig()
        p = AX.PacketControl(sock, mjig)
        p.sessionID = 42
        # Test successful
        pkt = AP.GetPDU(True, 42, 23, 100, [AP.SearchRange((0, 0), ())])
        p.handle_GetPDU(pkt)
        self.assertEqual(len(sock.data), 1)
        sentPkt = AP.decode_packet(sock.data[0])[0]
        respPkt = AP.ResponsePDU(True, 42, 23, 100, 0, 0, 0,
                                 (AP.Varbind(AP.VALUE_INTEGER, (0, 0), 1),))
        self.assertEqual(sentPkt, respPkt)
        # Test non-existent
        sock.data = []
        pkt = AP.GetPDU(True, 42, 23, 100, [AP.SearchRange((1, 0), ())])
        p.handle_GetPDU(pkt)
        self.assertEqual(len(sock.data), 1)
        sentPkt = AP.decode_packet(sock.data[0])[0]
        respPkt = AP.ResponsePDU(True, 42, 23, 100, 0, 0, 0,
                                 (AP.Varbind(AP.VALUE_NO_SUCH_OBJECT,
                                             (1, 0)),))
        self.assertEqual(sentPkt, respPkt)
        # Test non-existent
        sock.data = []
        pkt = AP.GetPDU(True, 42, 23, 100, [AP.SearchRange((0, 1), ())])
        p.handle_GetPDU(pkt)
        self.assertEqual(len(sock.data), 1)
        sentPkt = AP.decode_packet(sock.data[0])[0]
        respPkt = AP.ResponsePDU(True, 42, 23, 100, 0, 0, 0,
                                 (AP.Varbind(AP.VALUE_NULL,
                                             (0, 1)),))
        self.assertEqual(sentPkt, respPkt)

    def test_handleGetNextPDU(self):
        readReturns = [1]
        read = (lambda oid: AP.Varbind(AP.VALUE_INTEGER,
                                       oid, readReturns.pop(0)))
        mjig = AX.MIBControl()
        mjig.addNode((0, 0), read)
        mjig.addNode((0, 1), read)
        sock = jigs.SocketJig()
        p = AX.PacketControl(sock, mjig)
        p.sessionID = 42
        # Test successful
        pkt = AP.GetNextPDU(True, 42, 23, 100, [AP.SearchRange((0, 0), ())])
        p.handle_GetNextPDU(pkt)
        self.assertEqual(len(sock.data), 1)
        sentPkt = AP.decode_packet(sock.data[0])[0]
        respPkt = AP.ResponsePDU(True, 42, 23, 100, 0, 0, 0,
                                 (AP.Varbind(AP.VALUE_INTEGER, (0, 1), 1),))
        self.assertEqual(sentPkt, respPkt)
        # Test non-existent
        sock.data = []
        pkt = AP.GetNextPDU(True, 42, 23, 100, [AP.SearchRange((1, 0), ())])
        p.handle_GetNextPDU(pkt)
        self.assertEqual(len(sock.data), 1)
        sentPkt = AP.decode_packet(sock.data[0])[0]
        respPkt = AP.ResponsePDU(True, 42, 23, 100, 0, 0, 0,
                                 (AP.Varbind(AP.VALUE_END_OF_MIB_VIEW,
                                             (1, 0)),))
        self.assertEqual(sentPkt, respPkt)

    def test_handle_GetBulkPDU(self):
        readReturns = [1, 2, 3, 4, 5, 6]
        read = (lambda oid: AP.Varbind(AP.VALUE_INTEGER,
                                       oid, readReturns.pop(0)))
        mjig = AX.MIBControl()
        # non-repeated
        mjig.addNode((0, 0), read)
        mjig.addNode((0, 1), read)  # shouldn't see
        mjig.addNode((1, 0), read)
        mjig.addNode((1, 1), read)  # shouldn't see
        # repeated
        mjig.addNode((2, 0), read)
        mjig.addNode((2, 1), read)
        mjig.addNode((2, 2), read)  # shouldn't see
        mjig.addNode((3, 0), read)
        mjig.addNode((3, 1), read)
        mjig.addNode((3, 2), read)  # shouldn't see
        mjig.addNode((4, 0), read)  # shouldn't see
        sock = jigs.SocketJig()
        p = AX.PacketControl(sock, mjig)
        p.sessionID = 42
        pkt = AP.GetBulkPDU(True, 42, 23, 100, 2, 2,
                            [AP.SearchRange((0,), (1,), True),
                             AP.SearchRange((1,), (2,), True),
                             AP.SearchRange((2, 0), (3,), True),
                             AP.SearchRange((3, 0), (), True)])
        p.handle_GetBulkPDU(pkt)
        self.assertEqual(len(sock.data), 1)
        sentPkt = AP.decode_packet(sock.data[0])[0]
        respPkt = AP.ResponsePDU(True, 42, 23, 100, 0, 0, 0,
                                 (AP.Varbind(AP.VALUE_INTEGER, (0, 0), 1),
                                  AP.Varbind(AP.VALUE_INTEGER, (1, 0), 2),
                                  AP.Varbind(AP.VALUE_INTEGER, (2, 0), 3),
                                  AP.Varbind(AP.VALUE_INTEGER, (2, 1), 4),
                                  AP.Varbind(AP.VALUE_INTEGER, (3, 0), 5),
                                  AP.Varbind(AP.VALUE_INTEGER, (3, 1), 6)))
        self.assertEqual(sentPkt, respPkt)

    def test_handle_TestSetPDU(self):
        mjig = AX.MIBControl()
        readReturns = []
        read = (lambda oid: AP.Varbind(AP.VALUE_INTEGER,
                                       oid, readReturns.pop(0)))
        writeCalls = []
        writeReturns = []
        write = (lambda a, vb: (writeCalls.append((a, vb)),
                                writeReturns.pop(0))[1])
        mjig.addNode((0,), read, write)
        mjig.addNode((1,), read)
        sock = jigs.SocketJig()
        p = AX.PacketControl(sock, mjig)
        p.sessionID = 42
        # Test success
        readReturns = [10]
        writeReturns = [AP.ERR_NOERROR]
        pkt = AP.TestSetPDU(True, 42, 23, 100,
                            [AP.Varbind(AP.VALUE_INTEGER, (0,), 16)])
        p.handle_TestSetPDU(pkt)
        self.assertEqual(len(sock.data), 1)
        sentPkt = AP.decode_packet(sock.data[0])[0]
        respPkt = AP.ResponsePDU(True, 42, 23, 100, 0, 0, 0)
        self.assertEqual(sentPkt, respPkt)
        self.assertEqual(writeCalls,
                         [("test", AP.Varbind(AP.VALUE_INTEGER, (0,), 16))])
        self.assertEqual(mjig.setVarbinds,
                         [AP.Varbind(AP.VALUE_INTEGER, (0,), 16)])
        self.assertEqual(mjig.setHandlers, [write])
        self.assertEqual(mjig.setUndoData,
                         [AP.Varbind(AP.VALUE_INTEGER, (0,), 10)])
        # Test failure, no access
        mjig.setVarbinds = []
        mjig.setHandlers = []
        mjig.setUndoData = []
        sock.data = []
        readReturns = [10]
        writeReturns = [AP.ERR_NOERROR]
        writeCalls = []
        pkt = AP.TestSetPDU(True, 42, 23, 100,
                            [AP.Varbind(AP.VALUE_INTEGER, (0, 1), 16)])
        p.handle_TestSetPDU(pkt)
        self.assertEqual(len(sock.data), 1)
        sentPkt = AP.decode_packet(sock.data[0])[0]
        respPkt = AP.ResponsePDU(True, 42, 23, 100, 0, 6, 0)
        self.assertEqual(sentPkt, respPkt)
        self.assertEqual(writeCalls, [])
        self.assertEqual(mjig.setVarbinds, [])
        self.assertEqual(mjig.setHandlers, [])
        self.assertEqual(mjig.setUndoData, [])
        # Test failure, not writable
        mjig.setVarbinds = []
        mjig.setHandlers = []
        mjig.setUndoData = []
        sock.data = []
        readReturns = [10]
        writeReturns = [AP.ERR_NOERROR]
        writeCalls = []
        pkt = AP.TestSetPDU(True, 42, 23, 100,
                            [AP.Varbind(AP.VALUE_INTEGER, (1,), 16)])
        p.handle_TestSetPDU(pkt)
        self.assertEqual(len(sock.data), 1)
        sentPkt = AP.decode_packet(sock.data[0])[0]
        respPkt = AP.ResponsePDU(True, 42, 23, 100, 0, 17, 0)
        self.assertEqual(sentPkt, respPkt)
        self.assertEqual(writeCalls, [])
        self.assertEqual(mjig.setVarbinds, [])
        self.assertEqual(mjig.setHandlers, [])
        self.assertEqual(mjig.setUndoData, [])

    def test_handle_CommitSetPDU(self):
        handlerCalls = []
        handlerReturns = [AP.ERR_NOERROR, AP.ERR_NOERROR]
        handler = (lambda a, vb: (handlerCalls.append((a, vb)),
                   handlerReturns.pop(0))[1])
        mjig = AX.MIBControl()
        mjig.setVarbinds = ["foo", "bar"]
        mjig.setHandlers = [handler, handler]
        sock = jigs.SocketJig()
        p = AX.PacketControl(sock, mjig)
        p.sessionID = 42
        # Test success
        pkt = AP.CommitSetPDU(True, 42, 23, 100)
        p.handle_CommitSetPDU(pkt)
        self.assertEqual(len(sock.data), 1)
        sentPkt = AP.decode_packet(sock.data[0])[0]
        respPkt = AP.ResponsePDU(True, 42, 23, 100, 0, 0, 0)
        self.assertEqual(sentPkt, respPkt)
        self.assertEqual(handlerCalls, [("commit", "foo"),
                                        ("commit", "bar")])
        # Test failure
        sock.data = []
        handlerCalls = []
        handlerReturns = [AP.ERR_NOERROR, AP.ERR_GENERR]
        p.handle_CommitSetPDU(pkt)
        self.assertEqual(len(sock.data), 1)
        sentPkt = AP.decode_packet(sock.data[0])[0]
        respPkt = AP.ResponsePDU(True, 42, 23, 100, 0, 5, 1)
        self.assertEqual(sentPkt, respPkt)
        self.assertEqual(handlerCalls, [("commit", "foo"),
                                        ("commit", "bar")])

    def test_handle_UndoSetPDU(self):
        handlerCalls = []
        handlerReturns = [AP.ERR_NOERROR, AP.ERR_NOERROR]
        handler = (lambda a, vb, ud: (handlerCalls.append((a, vb, ud)),
                                      handlerReturns.pop(0))[1])
        mjig = AX.MIBControl()
        mjig.setVarbinds = ["foo", "bar"]
        mjig.setHandlers = [handler, handler]
        mjig.setUndoData = ["fey", "yar"]
        sock = jigs.SocketJig()
        p = AX.PacketControl(sock, mjig)
        p.sessionID = 42
        # Test success
        pkt = AP.UndoSetPDU(True, 42, 23, 100)
        p.handle_UndoSetPDU(pkt)
        self.assertEqual(len(sock.data), 1)
        sentPkt = AP.decode_packet(sock.data[0])[0]
        respPkt = AP.ResponsePDU(True, 42, 23, 100, 0, 0, 0)
        self.assertEqual(sentPkt, respPkt)
        self.assertEqual(handlerCalls, [("undo", "foo", "fey"),
                                        ("undo", "bar", "yar")])
        # Test failure
        sock.data = []
        handlerCalls = []
        handlerReturns = [AP.ERR_NOERROR, AP.ERR_GENERR]
        p.handle_UndoSetPDU(pkt)
        self.assertEqual(len(sock.data), 1)
        sentPkt = AP.decode_packet(sock.data[0])[0]
        respPkt = AP.ResponsePDU(True, 42, 23, 100, 0, 5, 1)
        self.assertEqual(sentPkt, respPkt)
        self.assertEqual(handlerCalls, [("undo", "foo", "fey"),
                                        ('undo', 'bar', 'yar')])

    def test_handle_CleanupSetPDU(self):
        handlerCalls = []
        handler = (lambda a, vb: handlerCalls.append((a, vb)))
        mjig = AX.MIBControl()
        mjig.setVarbinds = ["foo", "bar"]
        mjig.setHandlers = [handler, handler]
        p = AX.PacketControl(None, mjig)
        p.sessionID = 42
        pkt = AP.CleanupSetPDU(True, 42, 23, 100)
        p.handle_CleanupSetPDU(pkt)
        self.assertEqual(p.database.inSetP, False)
        self.assertEqual(handlerCalls, [("clean", "foo"), ("clean", "bar")])

    def test_handle_ResponsePDU(self):
        cbreturn = []

        def callback(pkt, origpkt):
            cbreturn.append((pkt, origpkt))
        opkt = AP.PingPDU(True, 42, 4, 50)
        incorrect_rpkt = AP.ResponsePDU(True, 23, 1, 99, 0, 0, 0)
        correct_rpkt = AP.ResponsePDU(True, 42, 4, 50, 0, 0, 0)
        sock = jigs.SocketJig()
        p = AX.PacketControl(sock, None)
        p.sessionID = 42
        p.packetLog = {(42, 4, 50): (30.0, opkt, callback)}
        # Test non-expected response (likely to happen in a timeout)
        p.handle_ResponsePDU(incorrect_rpkt)
        self.assertEqual(p.packetLog, {(42, 4, 50): (30.0, opkt, callback)})
        # Test expected response
        p.handle_ResponsePDU(correct_rpkt)
        self.assertEqual(p.packetLog, {})


class TestAgentx(unittest.TestCase):
    def test_walkMIBTree(self):
        f = AX.walkMIBTree

        # Test empty tree
        self.assertEqual(tuple(f({})), ())
        # Test flat, fully static tree
        self.assertEqual(tuple(f({0: {"reader": None, "writer": None,
                                      "subids": None},
                                  1: {},
                                  2: {"reader": None, "writer": None,
                                      "subids": None},
                                  5: {"reader": None, "writer": None,
                                      "subids": None}})),
                         ((AP.OID((0,)), None, None),
                          (AP.OID((1,)), None, None),
                          (AP.OID((2,)), None, None),
                          (AP.OID((5,)), None, None)))
        # Test nested, fully static tree
        self.assertEqual(tuple(f({0: {"reader": None, "writer": None,
                                      "subids": None},
                                  1: {"reader": None, "writer": None,
                                      "subids":
                                      {0: {"reader": None, "writer": None,
                                           "subids": None},
                                       1: {"reader": None, "writer": None,
                                           "subids":
                                           {42: {"reader": None,
                                                 "writer": None,
                                                 "subids": None}}}}},
                                  5: {"reader": None, "writer": None,
                                      "subids": None}})),
                         ((AP.OID((0,)), None, None),
                          (AP.OID((1,)), None, None),
                          (AP.OID((1, 0)), None, None),
                          (AP.OID((1, 1)), None, None),
                          (AP.OID((1, 1, 42)), None, None),
                          (AP.OID((5,)), None, None)))
        # Test nested, fully static tree, with rootpath
        self.assertEqual(tuple(f({0: {"reader": None, "writer": None,
                                      "subids": None},
                                  1: {"reader": None, "writer": None,
                                      "subids":
                                      {0: {"reader": None, "writer": None,
                                           "subids": None},
                                       1: {"reader": None, "writer": None,
                                           "subids":
                                           {42: {"reader": None,
                                                 "writer": None,
                                                 "subids": None}}}}},
                                  5: {"reader": None, "writer": None,
                                      "subids": None}},
                                 (23,))),
                         ((AP.OID((23, 0)), None, None),
                          (AP.OID((23, 1)), None, None),
                          (AP.OID((23, 1, 0)), None, None),
                          (AP.OID((23, 1, 1)), None, None),
                          (AP.OID((23, 1, 1, 42)), None, None),
                          (AP.OID((23, 5)), None, None)))
        # subid lambda for dynamic tree testing
        submaker = (lambda: {0: {"reader": None, "writer": None,
                                 "subids": None},
                             1: {"reader": None, "writer": None,
                                 "subids":
                                 {0: {"reader": None, "writer": None,
                                      "subids": None}}},
                             2: {"reader": None, "writer": None,
                                 "subids": None}})
        # Test tree with dynamic nodes
        self.assertEqual(tuple(f({0: {"reader": None, "writer": None,
                                      "subids": None},
                                  1: {"reader": None, "writer": None,
                                      "subids": submaker},
                                  2: {"reader": None, "writer": None,
                                      "subids": None}})),
                         ((AP.OID((0,)), None, None),
                          (AP.OID((1,)), None, None),
                          (AP.OID((1, 0)), None, None),
                          (AP.OID((1, 1)), None, None),
                          (AP.OID((1, 1, 0)), None, None),
                          (AP.OID((1, 2)), None, None),
                          (AP.OID((2,)), None, None)))
        # Test tree with dynamic nodes and root path
        self.assertEqual(tuple(f({0: {"reader": None, "writer": None,
                                      "static": True, "subids": None},
                                  1: {"reader": None, "writer": None,
                                      "static": False, "subids": submaker},
                                  2: {"reader": None, "writer": None,
                                      "static": True, "subids": None}},
                                 (23,))),
                         ((AP.OID((23, 0)), None, None),
                          (AP.OID((23, 1)), None, None),
                          (AP.OID((23, 1, 0)), None, None),
                          (AP.OID((23, 1, 1)), None, None),
                          (AP.OID((23, 1, 1, 0)), None, None),
                          (AP.OID((23, 1, 2)), None, None),
                          (AP.OID((23, 2)), None, None)))


if __name__ == '__main__':
    unittest.main()



#! @PYSHEBANG@
# -*- coding: utf-8 -*-

# Copyright the NTPsec project contributors
#
# SPDX-License-Identifier: BSD-2-Clause

"""\
usage: ntplogtemp [-h] [-l LOGFILE] [-o] [-q] [-v] [-w WAIT] [-V]

Program to log system temperatures

optional arguments:
  -h, --help            show this help message and exit
  -l LOGFILE, --logfile LOGFILE
                        append log data to LOGFILE instead of stdout
  -o, --once            Run the output once and exit
  -q, --quiet           be quite
  -v, --verbose         be verbose
  -w WAIT, --wait WAIT  Set delay time in seconds, default is 60
  -V, --version         show program's version number and exit

See the manual page for details.
"""

from __future__ import print_function, division

import argparse
import glob
import logging
import logging.handlers
import os
import re
import subprocess
import sys
import time


class logfile_header_class(logging.handlers.TimedRotatingFileHandler):
    'A class to modify the file logging handler.'
    def doRollover(self):
        'function to add header to new file on rotation.'
        if str is bytes:
            super(logfile_header_class, self).doRollover()
        else:
            super().doRollover()
        self.stream.write('# time, sensor, value\n')


def run_binary(cmd):
    """\
Run a binary
Return output if good, None if bad
"""

    try:
        # sadly subprocess.check_output() is not in Python 2.6
        # so use Popen()
        # this throws an exception if not found
        proc = subprocess.Popen(cmd,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT,
                                universal_newlines=True)
        output = proc.communicate()[0].split("\n")

        if proc.returncode:
            # non-zero return code, fail
            return None

    except:
        if args.verbose:
            sys.stderr.write("Unable to run %s binary\n" % cmd[0])
        # throws exception
        return None
    return output


class CpuTemp:
    "Sensors on the CPU Core"
    has_sensors = False

    def __init__(self):
        # check for sensors binary
        ret = run_binary(["sensors", "-h"])
        if ret is not None:
            self.has_sensors = True

        # pattern that matches the string that has the cpu temp
        self._pattern = re.compile(r'^\s+temp\d+_input:\s+([\d\.]+).*$')

    def get_data(self):
        "Collects the data and return the output as an array"
        if not self.has_sensors:
            return None

        _index = 0
        _data = []
        # grab the needed output
        output = run_binary(["sensors", "-u"])

        if output is not None:
            for record in output:
                match = self._pattern.match(record)
                if match and match.group(1):
                    _now = int(time.time())
                    _cpu_temperature = match.group(1)
                    _data.append('%d LM%s %s' % (_now, _index,
                                                 _cpu_temperature))
                    _index += 1
        else:
            self.has_sensors = False
            if args.verbose:
                sys.stderr.write("No sensors returned temperatures. ",
                                 "Have you run sensors-detect?\n")

        return _data


class SmartCtl:
    "Sensor on the Hard Drive"
    _drives = []
    has_smartctl = False

    def __init__(self):
        ret = run_binary(["smartctl", "-h"])
        if ret is not None:
            self.has_smartctl = True

        if self.has_smartctl:
            # Which drive to watch
            for child in glob.glob('/dev/sd?'):
                self._drives.append(child)
            for child in glob.glob('/dev/nvme?n?'):
                self._drives.append(child)
            self._drives = sorted(self._drives)

    def get_data(self):
        "Collects the data and return the output as an array"
        if not self.has_smartctl:
            return None

        data = []
        for _device in self._drives[:]:
            output = run_binary(["smartctl", "-A", _device])

            if output is None:
                # do not keep trying on failure
                self._drives.remove(_device)
            else:
                for line in output:
                    if line.startswith('194 '):
                        now = int(time.time())
                        temp = line.split()[9]
                        data.append('%d %s %s' % (now, _device, temp))
                        break
                    if line.startswith('Temperature:'):
                        now = int(time.time())
                        temp = line.split()[1]
                        data.append('%d %s %s' % (now, _device, temp))
                        break
        return data


class Temper:
    """\
    Reads 'temper-poll -c' for room temperature data.

    Before you can use this class you must have a TEMPer USB thermometer
    plugged in, and the temper-python package must be installed and configured.
    See their documentation for that procedure.
"""
    has_temper = False

    def __init__(self):
        # check for sensors binary
        ret = run_binary(["temper-poll", "-h"])
        if ret is not None:
            self.has_temper = True

    def get_data(self):
        "Collects the data and return the output as an array"
        if not self.has_temper:
            return None

        data = []
        _device = 'TEMPER0'
        # only one device can read the TEMPer at a time
        # collisions will happen, so retry a few times
        for attempt in range(0, 3):
            # grab the needed output
            output = run_binary(["temper-poll", "-c"])
            try:
                # make sure it is a temperature
                temp = float(output[0])
                now = int(time.time())
                data.append('%d %s %s' % (now, _device, temp))
                break
            except (TypeError, ValueError):
                # bad data, ignore it, for a bit
                if args.verbose:
                    sys.stderr.write("TEMPer-poll failed\n")

        if not data:
            self.has_temper = False
        return data


class ZoneTemp:
    "Zone sensors"

    def __init__(self):
        base_dir = '/sys/class/thermal/thermal_zone?/'
        self.zones = []
        for child in glob.glob(base_dir):
            self.zones.append(child)

    def get_data(self):
        "Collects the data and return the output as an array"
        _zone = 0
        _data = []
        for zone in self.zones:
            try:
                with open(os.path.join(zone, 'temp')) as _zone_data:
                    for line in _zone_data:
                        temp = float(line) / 1000
                        _now = int(time.time())
                        _data.append('%d ZONE%s %s' % (_now, _zone, temp))
                        _zone += 1
            except OSError:
                # temp might not be readable, if so continue on to the next one
                continue
        return _data


# Work with argvars
parser = argparse.ArgumentParser(description="Temperature sensor daemon",
                                 epilog="""See the manual page for details.""")
parser.add_argument('-l', '--logfile',
                    dest='logfile',
                    help="append log data to LOGFILE instead of stdout",
                    nargs=1)
parser.add_argument('-o', '--once',
                    dest='once',
                    help="Run the output once and exit",
                    action='store_true')
parser.add_argument('-q', '--quiet',
                    action="store_true",
                    dest='quiet',
                    help="be quite")
parser.add_argument('-v', '--verbose',
                    action="store_true",
                    dest='verbose',
                    help="be verbose")
parser.add_argument('-w', '--wait',
                    default=[60],
                    dest='wait',
                    help="Set delay time in seconds, default is 60",
                    nargs=1,
                    type=int)
parser.add_argument('-V', '--version',
                    action="version",
                    version="ntplogtemp ntpsec-@NTPSEC_VERSION_EXTENDED@")
args = parser.parse_args()


def logging_setup():
    "Create logging object"
    logFormat = logging.Formatter('%(message)s')
    # Create logger for cpuTemp
    tempLogger = logging.getLogger()
    tempLogger.setLevel(logging.INFO)
    # Create file handler
    if args.logfile:
        # log to logfile
        file = logfile_header_class(
            args.logfile[0],
            utc=True,
            when='midnight',
            interval=1)
    else:
        # log to stdout
        file = logging.StreamHandler(sys.stdout)

    file.setLevel(logging.INFO)
    # Create the formatter and add it to the handler
    file.setFormatter(logFormat)
    # Add the handler to the logger
    tempLogger.addHandler(file)
    return tempLogger


def logData(log, data):
    "log the data"
    if data is not None:
        for _item in data:
            log.info(_item)


def log_data():
    "Write all temperature readings to one file"
    # Create objects
    cpu = CpuTemp()
    zone = ZoneTemp()
    hdd = SmartCtl()
    temper = Temper()

    # Create the logger instance
    Logger = logging_setup()

    # Create data layout
    logData(Logger, ["# time, sensor, value"])

    # Write data to their respective logs
    while True:
        logData(Logger, zone.get_data())
        logData(Logger, cpu.get_data())
        logData(Logger, hdd.get_data())
        logData(Logger, temper.get_data())
        if args.once:
            sys.exit(0)
        time.sleep(args.wait[0])


args = parser.parse_args()
if os.getuid():
    sys.stderr.write("You must be root!\n")
    sys.exit(1)

try:
    log_data()
except KeyboardInterrupt:
    print("")    # be nice to bash



#! /usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright the NTPsec project contributors
#
# SPDX-License-Identifier: BSD-2-Clause

'''keygone.py -- A cruddy improvised NTPsec keyfile generator

The number in the group_iterator range is the number of keys to generate
per key type.

key_number starts as the first key number and is incremented.

max_bytes is the implementation-specific maximum number of entropy bytes
allowed per key.

the variables list_* are lists of lists. the inner lists store the output
length of the algorithm (block size?) and the standard form of the
algorithms name (in lowercase). *_bad list algorithms are believed
broken, *_good ones are not. This list believed accurate as of the
6th of May 2020.
'''
import argparse
import os
import secrets
import stat
import sys

list_md_bad = [
    [16, 16, 'md5'],

    [20, 16, 'sha1'],
    [20, 16, 'rmd160']
]
list_md_good = [
    [16, 16, 'sm4'],

    [28, 16, 'sha224'],
    [28, 16, 'sha3-224'],

    [32, 16, 'blake2s256'],
    [32, 16, 'sha256'],
    [32, 16, 'sha3-256'],

    [48, 16, 'sha384'],
    [48, 16, 'sha3-384'],

    [64, 16, 'blake2b512'],
    [64, 16, 'sha512'],
    [64, 16,  'sha3-512']
]
list_cmac_bad = [
]
list_cmac_good = [
    [16, 16, 'aes-128'],
    [16, 16, 'aria-128'],
    [16, 16, 'camellia-128'],

    [24, 24, 'aes-192'],
    [24, 24, 'aria-192'],
    [24, 24, 'camellia-192'],

    [32, 32, 'aes-256'],
    [32, 32, 'aria-256'],
    [32, 32, 'camellia-256'],
]


class KeyGone():
    'Set up, generate and output keys for NTPsec.'

    def __init__(self, start: int, gap: int):
        'Set up key generation for NTPsec.'
        self.gap = min(0, gap)
        self.index = start
        self.backing = {}

    def __str__(self):
        'Return a string containing the generated keys.'
        _ = ''
        for row in self.backing:
            _ += '%5d\t%13s\t%s\n' % (row, *self.backing[row])
        return _

    def to_file(self, oname: str):
        'Write the current keys to a file.'
        with open(oname, 'w') as of:
            of.write(str(self))

    def do_link(self, oname: str, link: str):
        'Write the current keys to a file and link to it elsewhere.'
        orig_umask = os.umask(stat.S_IWGRP | stat.S_IRWXO)
        os.umask(orig_umask)
        self.to_file(oname)
        if os.path.exists(link):
            os.remove(link)    # The symlink() line below matters
        os.symlink(oname, link)

    def add(self, algo: list, keys: int, hexed: bool = False):
        'Generate a slew of new keys according to specs.'
        length, mlength, algor = algo
        newlength = min(length, 32 if hexed else 20)  # Yuck, hardcoded magic
        if mlength > newlength:
            sys.stderr.write('"%s" excluded because minimum length %d exceeds truncation %d\n' % (algor, mlength, newlength))
            return
        for _ in range(keys):
            self.backing[self.index] = [algor, self._gen_key(newlength, hexed)]
            self.index += 1
        self.index += self.gap

    def _gen_key(self, length: int, hexed: bool) -> str:
        'Generate a single key.'
        if hexed:
            return secrets.token_hex(length)
        result = ''
        for _ in range(length):
            result += chr(0x24 + secrets.randbelow(0x5a))
        return result


def list_algos():
    'List the available algorithms by buckets.'
    chunks = []
    iterable = (('bad CMAC algos:', list_cmac_bad),
                ('good CMAC algos:', list_cmac_good),
                ('bad digest algos:', list_md_bad),
                ('good digest algos:', list_md_good)
                )
    for label, things in iterable:
        _ = label + '\n'
        for thing in things:
            _ += '%3dbit\t%s\n' % (thing[0] * 8, thing[1])
        chunks.append(_)
    sys.stderr.write('\n'.join(chunks))
    sys.exit(0)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='NTPsec key generator.')
    parser.add_argument('-L', '--list', action='store_true',
                        default=False, help='List known algorithms')
    parser.add_argument('-d', '--delete', nargs='+',
                        help='delete algorithm (repeatable) or "everything"')
    parser.add_argument('-a', '--add', nargs='+',
                        help='delete algorithm (repeatable) or "everything"')
    parser.add_argument('-f', '--file', help='Output to a file defaults to "ntp.keygone"',
                        default='ntp.keygone')
    parser.add_argument(
        '-s', '--link', help='create a symlink (requires file)')
    parser.add_argument('-c', '--console', action='store_true',
                        default=False, help='output to console')
    parser.add_argument('-n', '--number', type=int,
                        default=10, help='number of keys per group')
    parser.add_argument('-i', '--initial', type=int,
                        default=1, help='number of initial key')
    parser.add_argument('-g', '--gap', type=int, default=0,
                        help='gap between subsequent groups')
    parser.add_argument('-t', '--text', dest='fmt', action='store_false',
                        default=False, help='generate text keys (default)')
    parser.add_argument('-x', '--hex', dest='fmt',
                        action='store_true', help='generate hexadecimal keys')
    args = parser.parse_args()
    if args.list:
        list_algos()
    kg = KeyGone(args.initial, args.gap)
    set_cur = set(map(lambda x: x[2], list_cmac_good + list_md_good))
    set_all = set(map(lambda x: x[2], list_cmac_bad + list_md_bad))
    set_all.update(set_cur)
    if isinstance(args.delete, str):
        args.delete = [args.delete]
    elif args.delete is None:
        args.delete = []
    for iteM in list(args.delete):
        item = iteM.lower()
        if item in set_cur:
            set_cur.remove(item)
        elif item == 'everything':
            set_cur = set()
        else:
            sys.stderr.write('"%s" not in algos' % item)
    if isinstance(args.add, str):
        args.add = [args.add]
    elif args.add is None:
        args.add = []
    for iteM in list(args.add):
        item = iteM.lower()
        if item in set_cur:
            sys.stderr.write('"%s" already in algos' % item)
        if item in set_all:
            set_cur.add(item)
        elif item == 'everything':
            set_cur = set_all.copy()
        else:
            sys.stderr.write('"%s" not in available algos' % item)
    algos = list_cmac_good + list_md_good + list_cmac_bad + list_md_bad
    fail = False
    if args.initial < 1:
        fail = True
        sys.stderr.write('"initial" key number must be a positive integer')
    if args.number < 1:
        fail = True
        sys.stderr.write('"number" of keys per group must be a positive integer')
    if args.gap < 0:
        fail = True
        sys.stderr.write('the "gap" between groups must be non-negative integer')
    if len(algos) < 1:
        fail = True
        sys.stderr.write('At least one algorithm must be specified')
    if args.initial + (args.gap + args.number) * len(algos) - args.gap > 65535:
        fail = True
        sys.stderr.write('End number must be less than 65,536')
    if fail:
        sys.exit(1)
    for algo in algos:
        if algo[2] in set_cur:
            kg.add(algo, args.number, args.fmt)
    if args.file is not None:
        if args.link is not None:
            kg.do_link(args.file, args.link)
        else:
            kg.to_file(args.file)
    if args.console:
        print(str(kg))



# -*- coding: utf-8 -*-

# Copyright the NTPsec project contributors
#
# SPDX-License-Identifier: BSD-2-Clause

from __future__ import print_function, division

import select
import time
import sys

try:
    import ntp.util
    import ntp.agentx_packet
    ax = ntp.agentx_packet
except ImportError as e:
    sys.stderr.write(
        "AgentX: can't find Python AgentX Packet library.\n")
    sys.stderr.write("%s\n" % e)
    sys.exit(1)


defaultTimeout = 30
pingTime = 60


def gen_next(generator):
    if str is bytes:  # Python 2
        return generator.next()
    else:  # Python 3
        return next(generator)


class MIBControl:
    def __init__(self, oidTree=None, mibRoot=(), rangeSubid=0, upperBound=None,
                 mibContext=None):
        self.oidTree = {}  # contains callbacks for the MIB
        if oidTree is not None:
            self.oidTree = oidTree
        # The undo system is only for the last operation
        self.inSetP = False  # Are we currently in the set procedure?
        self.setVarbinds = []  # Varbind of the current set operation
        self.setHandlers = []  # Handlers for commit/undo/cleanup of set
        self.setUndoData = []  # Previous values for undoing
        self.mibRoot = mibRoot
        self.rangeSubid = rangeSubid
        self.upperBound = upperBound
        self.context = mibContext

    def mib_rootOID(self):
        return self.mibRoot

    def mib_rangeSubid(self):
        return self.rangeSubid

    def mib_upperBound(self):
        return self.upperBound

    def mib_context(self):
        return self.context

    def addNode(self, oid, reader=None, writer=None, dynamic=None):
        if isinstance(oid, ax.OID):  # get it in a mungable format
            oid = tuple(oid.subids)
        # dynamic is the generator for tables
        currentLevel = self.oidTree
        remainingOID = oid
        while True:
            node, remainingOID = ntp.util.slicedata(remainingOID, 1)
            node = node[0]
            if node not in currentLevel.keys():
                currentLevel[node] = {"reader": None, "writer": None,
                                      "subids": None}
            if not remainingOID:  # We have reached the target node
                currentLevel[node]["reader"] = reader
                currentLevel[node]["writer"] = writer
                if dynamic is not None:
                    # can't be both dynamic and non-dynamic
                    currentLevel[node]["subids"] = dynamic
                return
            else:
                if currentLevel[node]["subids"] is None:
                    currentLevel[node]["subids"] = {}
                currentLevel = currentLevel[node]["subids"]

    def getOID_core(self, nextP, searchoid, returnGenerator=False):
        gen = walkMIBTree(self.oidTree, self.mibRoot)
        while True:
            try:
                oid, reader, writer = gen_next(gen)
                if nextP:  # GetNext
                    # For getnext any OID greater than the start qualifies
                    oidhit = (oid > searchoid)
                else:  # Get
                    # For get we need a *specific* OID
                    oidhit = (oid.subids == searchoid.subids)
                if oidhit and (reader is not None):
                    # We only return OIDs that have a minimal implementation
                    # walkMIBTree handles the generation of dynamic trees
                    if returnGenerator:
                        return oid, reader, writer, gen
                    else:
                        return oid, reader, writer
            except StopIteration:  # Couldn't find anything in the tree
                if returnGenerator:
                    return None, None, None, None
                else:
                    return None, None, None

    # These exist instead of just using getOID_core so semantics are clearer
    def getOID(self, searchoid, returnGenerator=False):
        "Get the requested OID"
        return self.getOID_core(False, searchoid, returnGenerator)

    def getNextOID(self, searchoid, returnGenerator=False):
        "Get the next lexicographical OID"
        return self.getOID_core(True, searchoid, returnGenerator)

    def getOIDsInRange(self, oidrange, firstOnly=False):
        "Get a list of every (optionally the first) OID in a range"
        oids = []
        gen = walkMIBTree(self.oidTree, self.mibRoot)
        # Find the first OID
        while True:
            try:
                oid, reader, writer = gen_next(gen)
                if reader is None:
                    continue  # skip unimplemented OIDs
                elif oid.subids == oidrange.start.subids:
                    # ok, found the start, do we need to skip it?
                    if oidrange.start.include:
                        oids.append((oid, reader, writer))
                        break
                    else:
                        continue
                elif oid > oidrange.start:
                    # If we are here it means we hit the start but skipped
                    if not oidrange.end.isNull() and oid >= oidrange.end:
                        # We fell off the range
                        return []
                    oids.append((oid, reader, writer))
                    break
            except StopIteration:
                # Couldn't find *anything*
                return []
        if firstOnly:
            return oids
        # Start filling in the rest of the range
        while True:
            try:
                oid, reader, writer = gen_next(gen)
                if reader is None:
                    continue  # skip unimplemented OIDs
                elif not oidrange.end.isNull() and oid >= oidrange.end:
                    break  # past the end of a bounded range
                else:
                    oids.append((oid, reader, writer))
            except StopIteration:
                break  # We have run off the end of the MIB
        return oids


class PacketControl:
    def __init__(self, sock, dbase, spinGap=0.001, timeout=defaultTimeout,
                 logfp=None, debug=10000):
        self.log = (lambda txt, dbg: ntp.util.dolog(logfp, txt, debug, dbg))
        # take a pre-made socket instead of making our own so that
        # PacketControl doesn't have to know or care about implementation
        self.socket = sock
        self.spinGap = spinGap  # sleep() time on each loop
        # indexed on: (session_id, transaction_id, packet_id)
        # contains: (timeout, packet class)
        self.packetLog = {}  # Sent packets kept until response is received
        self.loopCallback = None  # called each loop in runforever mode
        self.database = dbase  # class for handling data requests
        self.receivedData = b""  # buffer for data from incomplete packets
        self.receivedPackets = []  # use as FIFO
        self.timeout = timeout
        self.sessionID = None  # need this for all packets
        self.highestTransactionID = 0  # used for exchanges we start
        self.lastReception = None
        self.stillConnected = False
        # indexed on pdu code
        self.pduHandlers = {ax.PDU_GET: self.handle_GetPDU,
                            ax.PDU_GET_NEXT: self.handle_GetNextPDU,
                            ax.PDU_GET_BULK: self.handle_GetBulkPDU,
                            ax.PDU_TEST_SET: self.handle_TestSetPDU,
                            ax.PDU_COMMIT_SET: self.handle_CommitSetPDU,
                            ax.PDU_UNDO_SET: self.handle_UndoSetPDU,
                            ax.PDU_CLEANUP_SET: self.handle_CleanupSetPDU,
                            ax.PDU_RESPONSE: self.handle_ResponsePDU}

    def mainloop(self, runforever):
        if self.stillConnected is not True:
            return False
        if runforever:
            while self.stillConnected:
                self._doloop()
                if self.loopCallback is not None:
                    self.loopCallback(self)
                time.sleep(self.spinGap)
        else:
            self._doloop()
        return self.stillConnected

    def _doloop(self):
        # loop body split out to separate the one-shot/run-forever switches
        # from the actual logic
        self.packetEater()
        while self.receivedPackets:
            packet = self.receivedPackets.pop(0)
            if packet.sessionID != self.sessionID:
                self.log(
                    "Received packet with incorrect session ID: %s" % packet,
                    3)
                resp = ax.ResponsePDU(True, packet.sessionID,
                                      packet.transactionID, packet.packetID,
                                      0, ax.RSPERR_NOT_OPEN, 0)
                self.sendPacket(resp, False)
                continue
            ptype = packet.pduType
            if ptype in self.pduHandlers:
                self.pduHandlers[ptype](packet)
            else:
                self.log("Dropping packet type %i, not implemented" % ptype,
                         2)
        self.checkResponses()
        if self.lastReception is not None:
            currentTime = time.time()
            if (currentTime - self.lastReception) > pingTime:
                self.sendPing()

    def initNewSession(self):
        self.log("Initializing new session...", 3)
        # We already have a connection, need to open a session.
        openpkt = ax.OpenPDU(True, 23, 0, 0, self.timeout, (),
                             "NTPsec SNMP subagent")
        self.sendPacket(openpkt, False)
        response = self.waitForResponse(openpkt, True)
        self.sessionID = response.sessionID
        # Register the tree
        register = ax.RegisterPDU(True, self.sessionID, 1, 1, self.timeout, 1,
                                  self.database.mib_rootOID(),
                                  self.database.mib_rangeSubid(),
                                  self.database.mib_upperBound(),
                                  self.database.mib_context())
        self.sendPacket(register, False)
        self.waitForResponse(register)
        self.stillConnected = True

    def waitForResponse(self, opkt, ignoreSID=False):
        "Wait for a response to a specific packet, dropping everything else"
        while True:
            self.packetEater()
            while self.receivedPackets:
                packet = self.receivedPackets.pop(0)
                if packet.__class__ != ax.ResponsePDU:
                    continue
                haveit = (opkt.transactionID == packet.transactionID) and \
                         (opkt.packetID == packet.packetID)
                if not ignoreSID:
                    haveit = haveit and (opkt.sessionID == packet.sessionID)
                if haveit:
                    self.log("Received waited for response", 4)
                    return packet
            time.sleep(self.spinGap)

    def checkResponses(self):
        "Check for expected responses that have timed out"
        currentTime = time.time()
        for key in list(self.packetLog.keys()):
            expiration, originalPkt, callback = self.packetLog[key]
            if currentTime > expiration:
                if callback is not None:
                    callback(None, originalPkt)
                del self.packetLog[key]

    def packetEater(self):
        "Slurps data from the input buffer and tries to parse packets from it"
        self.pollSocket()
        while True:
            datalen = len(self.receivedData)
            if datalen < 20:
                return None  # We don't even have a packet header, bail
            try:
                pkt, fullPkt, extraData = ax.decode_packet(self.receivedData)
                if not fullPkt:
                    return None
                self.receivedData = extraData
                self.receivedPackets.append(pkt)
                if pkt.transactionID > self.highestTransactionID:
                    self.highestTransactionID = pkt.transactionID
                self.log("Received a full packet: %s" % repr(pkt), 4)
            except (ax.ParseVersionError, ax.ParsePDUTypeError,
                    ax.ParseError) as e:
                if e.header["type"] != ax.PDU_RESPONSE:
                    # Response errors are silently dropped, per RFC
                    # Everything else sends an error response
                    self.sendErrorResponse(e.header, ax.RSPERR_PARSE_ERROR, 0)
                # *Hopefully* the packet length was correct.....
                #  if not, all packets will be scrambled. Maybe dump the
                #  whole buffer if too many failures in a row?
                self.receivedData = e.remainingData

    def sendPacket(self, packet, expectsReply, replyTimeout=defaultTimeout,
                   callback=None):
        encoded = packet.encode()
        self.log("Sending packet (with reply: %s): %s" % (expectsReply,
                                                          repr(packet)), 4)
        self.socket.sendall(encoded)
        if expectsReply:
            index = (packet.sessionID,
                     packet.transactionID,
                     packet.packetID)
            self.packetLog[index] = (replyTimeout, packet, callback)

    def sendPing(self):
        # DUMMY packetID, does this need to change? or does the pktID only
        # count relative to a given transaction ID?
        tid = self.highestTransactionID + 5  # +5 to avoid collisions
        self.highestTransactionID = tid
        pkt = ax.PingPDU(True, self.sessionID, tid, 1)

        def callback(resp, orig):
            if resp is None:  # Timed out. Need to restart the session.
                # Er, problem: Can't handle reconnect from inside PacketControl
                self.stillConnected = False
        self.sendPacket(pkt, True, callback=callback)

    def sendNotify(self, varbinds, context=None):
        # DUMMY packetID, does this need to change? or does the pktID only
        # count relative to a given transaction ID?
        tid = self.highestTransactionID + 5  # +5 to avoid collisions
        self.highestTransactionID = tid
        pkt = ax.NotifyPDU(True, self.sessionID, tid, 1, varbinds, context)

        def resendNotify(pkt, orig):
            if pkt is None:
                self.sendPacket(orig, True, callback=resendNotify)
        self.sendPacket(pkt, True, resendNotify)

    def sendErrorResponse(self, errorHeader, errorType, errorIndex):
        err = ax.ResponsePDU(errorHeader["flags"]["bigEndian"],
                             errorHeader["session_id"],
                             errorHeader["transaction_id"],
                             errorHeader["packet_id"],
                             0, errorType, errorIndex)
        self.sendPacket(err, False)

    def pollSocket(self):
        "Reads all currently available data from the socket, non-blocking"
        data = b""
        while True:
            tmp = select.select([self.socket], [], [], 0)[0]
            if not tmp:  # No socket, means no data available
                break
            tmp = tmp[0]
            newdata = tmp.recv(4096)  # Arbitrary value
            if newdata:
                self.log("Received data: %s" % repr(newdata), 5)
                data += newdata
                self.lastReception = time.time()
            else:
                break
        self.receivedData += data

    # ==========================
    # Packet handlers start here
    # ==========================

    def handle_GetPDU(self, packet):
        binds = []
        for oidr in packet.oidranges:
            target = oidr.start
            oid, reader, _ = self.database.getOID(target)
            if (oid != target) or (reader is None):
                # This OID must not be implemented yet.
                binds.append(ax.Varbind(ax.VALUE_NO_SUCH_OBJECT, target))
            else:
                vbind = reader(oid)
                if vbind is None:  # No data available.
                    # I am not certain that this is the correct response
                    # when no data is available. snmpwalk appears to stop
                    # calling a particular sub-agent when it gets to a NULL.
                    binds.append(ax.Varbind(ax.VALUE_NULL, target))
                else:
                    binds.append(vbind)
            # There should also be a situation that leads to noSuchInstance
            #  but I do not understand the requirements for that
        # TODO: Need to implement genError
        resp = ax.ResponsePDU(True, self.sessionID, packet.transactionID,
                              packet.packetID, 0, ax.ERR_NOERROR, 0, binds)
        self.sendPacket(resp, False)

    def handle_GetNextPDU(self, packet):
        binds = []
        for oidr in packet.oidranges:
            while True:
                oids = self.database.getOIDsInRange(oidr, True)
                if not oids:  # Nothing found
                    binds.append(ax.Varbind(ax.VALUE_END_OF_MIB_VIEW,
                                            oidr.start))
                    break
                else:
                    oid, reader, _ = oids[0]
                    vbind = reader(oid)
                    if vbind is None:  # No data available
                        # Re-do search for this OID range, starting from just
                        # after the current location
                        oidr = ax.SearchRange(oid, oidr.end, False)
                        continue
                    else:
                        binds.append(vbind)
                        break
        # TODO: Need to implement genError
        resp = ax.ResponsePDU(True, self.sessionID, packet.transactionID,
                              packet.packetID, 0, ax.ERR_NOERROR, 0, binds)
        self.sendPacket(resp, False)

    def handle_GetBulkPDU(self, packet):
        binds = []
        nonreps = packet.oidranges[:packet.nonReps]
        repeats = packet.oidranges[packet.nonReps:]
        # Handle non-repeats
        for oidr in nonreps:
            oids = self.database.getOIDsInRange(oidr, True)
            if not oids:  # Nothing found
                binds.append(ax.Varbind(ax.VALUE_END_OF_MIB_VIEW, oidr.start))
            else:
                oid, reader, _ = oids[0]
                binds.append(reader(oid))
        # Handle repeaters
        for oidr in repeats:
            oids = self.database.getOIDsInRange(oidr)
            if not oids:  # Nothing found
                binds.append(ax.Varbind(ax.VALUE_END_OF_MIB_VIEW, oidr.start))
            else:
                for oid, reader, _ in oids[:packet.maxReps]:
                    binds.append(reader(oid))
        resp = ax.ResponsePDU(True, self.sessionID, packet.transactionID,
                              packet.packetID, 0, ax.ERR_NOERROR, 0, binds)
        self.sendPacket(resp, False)

    def handle_TestSetPDU(self, packet):  # WIP / TODO
        # Be advised: MOST OF THE VALIDATION IS DUMMY CODE OR DOESN'T EXIST
        # According to the RFC this is one of the most demanding parts and
        #  *has* to be gotten right
        if self.database.inSetP:
            pass  # Is this an error?
        # if (inSetP) is an error these will go in an else block
        self.database.inSetP = True
        self.database.setVarbinds = []
        self.database.setHandlers = []
        self.database.setUndoData = []
        error = None
        for bindIndex in range(len(packet.varbinds)):
            varbind = packet.varbinds[bindIndex]
            # Find an OID, then validate it
            oid, reader, writer = self.database.getOID(varbind.oid)
            if oid is None:  # doesn't exist, can we create it?
                # DUMMY, assume we can't create anything
                error = ax.ERR_NO_ACCESS
                break
            elif writer is None:  # exists, writing not implemented
                error = ax.ERR_NOT_WRITABLE
                break
            # Ok, we have an existing or new OID, assemble the orders
            # If we created a new bind undoData is None, must delete it
            undoData = reader(oid)
            error = writer("test", varbind)
            if error != ax.ERR_NOERROR:
                break
            self.database.setVarbinds.append(varbind)
            self.database.setHandlers.append(writer)
            self.database.setUndoData.append(undoData)
        if error != ax.ERR_NOERROR:
            resp = ax.ResponsePDU(True, self.sessionID, packet.transactionID,
                                  packet.packetID, 0, error, bindIndex)
            self.sendPacket(resp, False)
            for i in range(bindIndex):
                # Errored out, clear the successful ones
                self.database.setHandlers[i]("clear",
                                             self.database.setVarbinds[i])
            self.database.inSetP = False
        else:
            resp = ax.ResponsePDU(True, self.sessionID, packet.transactionID,
                                  packet.packetID, 0, ax.ERR_NOERROR, 0)
            self.sendPacket(resp, False)

    def handle_CommitSetPDU(self, packet):
        if not self.database.inSetP:
            pass  # how to handle this?
        varbinds = self.database.setVarbinds
        handlers = self.database.setHandlers
        for i in range(len(varbinds)):
            error = handlers[i]("commit", varbinds[i])
            if error != ax.ERR_NOERROR:
                break
        if error != ax.ERR_NOERROR:
            resp = ax.ResponsePDU(True, self.sessionID, packet.transactionID,
                                  packet.packetID, 0, error, i)
        else:
            resp = ax.ResponsePDU(True, self.sessionID, packet.transactionID,
                                  packet.packetID, 0, ax.ERR_NOERROR, 0)
        self.sendPacket(resp, False)

    def handle_UndoSetPDU(self, packet):
        varbinds = self.database.setVarbinds
        handlers = self.database.setHandlers
        undoData = self.database.setUndoData
        for i in range(len(varbinds)):
            error = handlers[i]("undo", varbinds[i], undoData[i])
            if error != ax.ERR_NOERROR:
                break
        if error != ax.ERR_NOERROR:
            resp = ax.ResponsePDU(True, self.sessionID, packet.transactionID,
                                  packet.packetID, 0, error, i)
        else:
            resp = ax.ResponsePDU(True, self.sessionID, packet.transactionID,
                                  packet.packetID, 0, ax.ERR_NOERROR, 0)
        self.sendPacket(resp, False)

    def handle_CleanupSetPDU(self, packet):
        varbinds = self.database.setVarbinds
        handlers = self.database.setHandlers
        for i in range(len(varbinds)):
            handlers[i]("clean", varbinds[i])
        self.database.inSetP = False

    def handle_ResponsePDU(self, packet):
        index = (packet.sessionID, packet.transactionID, packet.packetID)
        if index in self.packetLog:
            timeout, originalPkt, callback = self.packetLog[index]
            del self.packetLog[index]
            if callback is not None:
                callback(packet, originalPkt)
        else:
            # Ok, response with no associated packet.
            # Probably something that timed out.
            pass


def walkMIBTree(tree, rootpath=()):
    # Tree node formats:
    # {"reader": <func>, "writer": <func>, "subids": {.blah.}}
    # {"reader": <func>, "writer": <func>, "subids": <func>}
    # The "subids" function in dynamic nodes must return an MIB tree
    nodeStack = []
    oidStack = []
    current = tree
    currentKeys = list(current.keys())
    currentKeys.sort()
    keyID = 0
    while True:
        if keyID >= len(currentKeys):
            if nodeStack:
                # No more nodes this level, pop higher node
                current, currentKeys, keyID, key = nodeStack.pop()
                oidStack.pop()
                keyID += 1
                continue
            else:  # Out of tree, we are done
                return
        key = currentKeys[keyID]
        oid = ax.OID(rootpath + tuple(oidStack) + (key,))
        yield (oid, current[key].get("reader"), current[key].get("writer"))
        subs = current[key].get("subids")
        if subs is not None:
            # Push current node, move down a level
            nodeStack.append((current, currentKeys, keyID, key))
            oidStack.append(key)
            if isinstance(subs, dict):
                current = subs
            else:
                current = subs()  # Tree generator function
                if current == {}:  # no dynamic subids, pop
                    current, currentKeys, keyID, key = nodeStack.pop()
                    oidStack.pop()
                    keyID += 1
                    continue
            currentKeys = list(current.keys())
            currentKeys.sort()
            keyID = 0
            key = currentKeys[keyID]
            continue
        keyID += 1



# -*- coding: utf-8 -*-
"Common utility functions"

# Copyright the NTPsec project contributors
#
# SPDX-License-Identifier: BSD-2-Clause

from __future__ import print_function, division


import collections
import os
import re
import shutil
import socket
import string
import sys
import time
import ntp.ntpc
import ntp.magic
import ntp.control


# Old CTL_PST defines for version 2.
OLD_CTL_PST_CONFIG = 0x80
OLD_CTL_PST_AUTHENABLE = 0x40
OLD_CTL_PST_AUTHENTIC = 0x20
OLD_CTL_PST_REACH = 0x10
OLD_CTL_PST_SANE = 0x08
OLD_CTL_PST_DISP = 0x04

OLD_CTL_PST_SEL_REJECT = 0
OLD_CTL_PST_SEL_SELCAND = 1
OLD_CTL_PST_SEL_SYNCCAND = 2
OLD_CTL_PST_SEL_SYSPEER = 3


# Units for formatting
UNIT_NS = "ns"       # nano second
UNIT_US = u"µs"      # micro second
UNIT_MS = "ms"       # milli second
UNIT_S = "s"         # second
UNIT_KS = "ks"       # kilo seconds
UNITS_SEC = [UNIT_NS, UNIT_US, UNIT_MS, UNIT_S, UNIT_KS]
UNIT_PPT = "ppt"     # parts per trillion
UNIT_PPB = "ppb"     # parts per billion
UNIT_PPM = "ppm"     # parts per million
UNIT_PPK = u"‰"      # parts per thousand
UNITS_PPX = [UNIT_PPT, UNIT_PPB, UNIT_PPM, UNIT_PPK]
unitgroups = (UNITS_SEC, UNITS_PPX)


# These two functions are not tested because they will muck up the module
# for everything else, and they are simple.

def check_unicode():  # pragma: no cover
    if "UTF-8" != sys.stdout.encoding:
        deunicode_units()
        return True  # needed by ntpmon
    return False


def deunicode_units():  # pragma: no cover
    """Under certain conditions it is not possible to force unicode output,
    this overwrites units that contain unicode with safe versions"""
    global UNIT_US
    global UNIT_PPK
    # Replacement units
    new_us = "us"
    new_ppk = "ppk"
    # Replace units in unit groups
    UNITS_SEC[UNITS_SEC.index(UNIT_US)] = new_us
    UNITS_PPX[UNITS_PPX.index(UNIT_PPK)] = new_ppk
    # Replace the units themselves
    UNIT_US = new_us
    UNIT_PPK = new_ppk


# Variables that have units
S_VARS = ("tai", "poll")
MS_VARS = ("rootdelay", "rootdisp", "rootdist", "offset", "sys_jitter",
           "clk_jitter", "leapsmearoffset", "authdelay", "koffset", "kmaxerr",
           "kesterr", "kprecis", "kppsjitter", "clk_wander_threshold",
           "tick", "in", "out", "bias", "delay", "jitter", "dispersion",
           "fudgetime1", "fudgetime2")
PPM_VARS = ("frequency", "clk_wander")


def dolog(logfp, text, debug, threshold):
    """debug is the current debug value
    threshold is the trigger for the current log"""
    if logfp is None:
        return  # can turn off logging by supplying a None file descriptor
    text = rfc3339(time.time()) + " " + text + "\n"
    if debug >= threshold:
        logfp.write(text)
        logfp.flush()  # we don't want to lose an important log to a crash


def safeargcast(arg, castfunc, errtext, usage):
    """Attempts to typecast an argument, prints and dies on failure.
    errtext must contain a %s for splicing in the argument, and be
    newline terminated."""
    try:
        casted = castfunc(arg)
    except ValueError:
        sys.stderr.write(errtext % arg)
        sys.stderr.write(usage)
        raise SystemExit(1)
    return casted


def stdversion():
    "Returns the NTPsec version string in a standard format"
    return "ntpsec-%s" % "@NTPSEC_VERSION_EXTENDED@"


def stdversioncheck(foreign):
    "Print a warning to stderr if module and foreign versions do not match."
    if stdversion() != foreign:
        sys.stderr.write("Module/Binary version mismatch\n")
        sys.stderr.write("Binary: %s\n" % foreign)
        sys.stderr.write("Module: %s\n" % ntp.util.stdversion())


def rfc3339(t):
    "RFC 3339 string from Unix time, including fractional second."
    rep = time.strftime("%Y-%m-%dT%H:%M:%S", time.gmtime(t))
    t = str(t)
    if "." in t:
        subsec = t.split(".", 1)[1]
        if int(subsec) > 0:
            rep += "." + subsec
    rep += "Z"
    return rep


def deformatNTPTime(txt):
    txt = txt[2:]  # Strip '0x'
    txt = "".join(txt.split("."))  # Strip '.'
    value = ntp.util.hexstr2octets(txt)
    return value


def hexstr2octets(hexstr):
    if (len(hexstr) % 2) != 0:
        hexstr = hexstr[:-1]  # slice off the last char
    values = []
    for index in range(0, len(hexstr), 2):
        values.append(chr(int(hexstr[index:index+2], 16)))
    return "".join(values)


def slicedata(data, slicepoint):
    "Breaks a sequence into two pieces at the slice point"
    return data[:slicepoint], data[slicepoint:]


def portsplit(hostname):
    portsuffix = ""
    if hostname.count(":") == 1:                # IPv4 with appended port
        (hostname, portsuffix) = hostname.split(":")
        portsuffix = ":" + portsuffix
    elif ']' in hostname:                       # IPv6
        rbrak = hostname.rindex("]")
        if ":" in hostname[rbrak:]:
            portsep = hostname.rindex(":")
            portsuffix = hostname[portsep:]
            hostname = hostname[:portsep]
            hostname = hostname[1:-1]   # Strip brackets
    return (hostname, portsuffix)


def parseConf(text):
    inQuote = False
    quoteStarter = ""
    lines = []
    tokens = []
    current = []

    def pushToken():
        token = "".join(current)
        if not inQuote:  # Attempt type conversion
            try:
                token = int(token)
            except ValueError:
                try:
                    token = float(token)
                except ValueError:
                    pass
        wrapper = (inQuote, token)
        tokens.append(wrapper)
        current[:] = []

    def pushLine():
        if current:
            pushToken()
        if tokens:
            lines.append(tokens[:])
            tokens[:] = []

    i = 0
    tlen = len(text)
    while i < tlen:
        if inQuote:
            if text[i] == quoteStarter:  # Ending a text string
                pushToken()
                quoteStarter = ""
                inQuote = False
            elif text[i] == "\\":  # Starting an escape sequence
                i += 1
                if text[i] in "'\"n\\":
                    current.append(eval("\'\\" + text[i] + "\'"))
            else:
                current.append(text[i])
        else:
            if text[i] == "#":  # Comment
                while (i < tlen) and (text[i] != "\n"):
                    i += 1  # Advance to end of line...
                i -= 1  # ...and back up so we don't skip the newline
            elif text[i] in "'\"":  # Starting a text string
                inQuote = True
                quoteStarter = text[i]
                if current:
                    pushToken()
            elif text[i] == "\\":  # Linebreak escape
                i += 1
                if text[i] != "\n":
                    raise SyntaxError
            elif text[i] == "\n":  # EOL: break the lines
                pushLine()
            elif text[i] in string.whitespace:
                if current:
                    pushToken()
            else:
                current.append(text[i])
        i += 1
    pushLine()
    return lines


def stringfilt(data):
    "Pretty print string of space separated numbers"
    parts = data.split()

    cooked = []
    for part in parts:
        # These are expected to fit on a single 80-char line.
        # Accounting for other factors this leaves each number with
        # 7 chars + a space.
        fitted = fitinfield(part, 7)
        cooked.append(fitted)
    rendered = " ".join(cooked)
    return rendered


def stringfiltcooker(data):
    "Cooks a filt* string of space separated numbers, expects milliseconds"
    parts = data.split()
    oomcount = {}
    minscale = -100000  # Keep track of the maxdownscale for each value
    # Find out what the 'natural' unit of each value is
    for part in parts:
        # Only care about OOMs, the real scaling happens later
        value, oom = scalestring(part)
        # Track the highest maxdownscale so we do not invent precision
        ds = maxdownscale(part)
        minscale = max(ds, minscale)
        oomcount[oom] = oomcount.get(oom, 0) + 1
    # Find the most common unit
    mostcommon = 0
    highestcount = 0
    for key in oomcount.keys():
        if key < minscale:
            continue  # skip any scale that would result in making up data
        count = oomcount[key]
        if count > highestcount:
            mostcommon = key
            highestcount = count
    # Shift all values to the new unit
    cooked = []
    for part in parts:
        part = rescalestring(part, mostcommon)
        fitted = fitinfield(part, 7)
        cooked.append(fitted)
    rendered = " ".join(cooked) + " " + UNITS_SEC[mostcommon +
                                                  UNITS_SEC.index(UNIT_MS)]
    return rendered


def getunitgroup(unit):
    "Returns the unit group which contains a given unit"
    for group in unitgroups:
        if unit in group:
            return group


def oomsbetweenunits(a, b):
    "Calculates how many orders of magnitude separate two units"
    group = getunitgroup(a)
    if b is None:  # Caller is asking for the distance from the base unit
        return group.index(a) * 3
    elif b in group:
        ia = group.index(a)
        ib = group.index(b)
        return abs((ia - ib) * 3)
    return None


def breaknumberstring(value):
    "Breaks a number string into (aboveDecimal, belowDecimal, isNegative?)"
    if value[0] == "-":
        value = value[1:]
        negative = True
    else:
        negative = False
    if "." in value:
        above, below = value.split(".")
    else:
        above = value
        below = ""
    return (above, below, negative)


def gluenumberstring(above, below, isnegative):
    "Glues together parts of a number string"
    if above == "":
        above = "0"
    if below:
        newvalue = ".".join((above, below))
    else:
        newvalue = above
    if isnegative:
        newvalue = "-" + newvalue
    return newvalue


def maxdownscale(value):
    "Maximum units a value can be scaled down without inventing data"
    if "." in value:
        digitcount = len(value.split(".")[1])
        # Return a negative so it can be fed directly to a scaling function
        return -(digitcount // 3)
    else:
        # No decimals, the value is already at the maximum down-scale
        return 0


def rescalestring(value, unitsscaled):
    "Rescale a number string by a given number of units"
    whole, dec, negative = breaknumberstring(value)
    if unitsscaled == 0:
        # This may seem redundant, but glue forces certain formatting details
        value = gluenumberstring(whole, dec, negative)
        return value
    hilen = len(whole)
    lolen = len(dec)
    digitsmoved = abs(unitsscaled * 3)
    if unitsscaled > 0:  # Scale to a larger unit, move decimal left
        if hilen < digitsmoved:
            # Scaling beyond the digits, pad it out. We can pad here
            # without making up digits that don't exist
            padcount = digitsmoved - hilen
            newwhole = ""
            newdec = ("0" * padcount) + whole + dec
        else:  # Scaling in the digits, no need to pad
            choppoint = -digitsmoved
            newdec = whole[choppoint:] + dec
            newwhole = whole[:choppoint]
    elif unitsscaled < 0:  # scale to a smaller unit, move decimal right
        if lolen < digitsmoved:
            # Scaling beyond the digits would force us to make up data
            # that doesn't exist. So fail.
            # The caller should have already caught this with maxdownscale()
            return None
        else:
            newwhole = whole + dec[:digitsmoved]
            newdec = dec[digitsmoved:]
    newwhole = newwhole.lstrip("0")
    newvalue = gluenumberstring(newwhole, newdec, negative)
    return newvalue


def formatzero(value):
    "Scale a zero value for the unit with the highest available precision"
    scale = maxdownscale(value)
    newvalue = rescalestring(value, scale).lstrip("-")
    return (newvalue, scale)


def scalestring(value):
    "Scales a number string to fit in the range 1.0-999.9"
    if isstringzero(value):
        return formatzero(value)
    whole, dec, negative = breaknumberstring(value)
    hilen = len(whole)
    if (hilen == 0) or isstringzero(whole):  # Need to shift to smaller units
        i = 0
        lolen = len(dec)
        while i < lolen:  # need to find the actual digits
            if dec[i] != "0":
                break
            i += 1
        lounits = (i // 3) + 1  # always need to shift one more unit
        movechars = lounits * 3
        if lolen < movechars:
            # Not enough digits to scale all the way down. Inventing
            # digits is unacceptable, so scale down as much as we can.
            lounits = (i // 3)  # "always", unless out of digits
            movechars = lounits * 3
        newwhole = dec[:movechars].lstrip("0")
        newdec = dec[movechars:]
        unitsmoved = -lounits
    else:  # Shift to larger units
        hiunits = hilen // 3  # How many we have, not how many to move
        hidigits = hilen % 3
        if hidigits == 0:  # full unit above the decimal
            hiunits -= 1  # the unit above the decimal doesn't count
            hidigits = 3
        newwhole = whole[:hidigits]
        newdec = whole[hidigits:] + dec
        unitsmoved = hiunits
    newvalue = gluenumberstring(newwhole, newdec, negative)
    return (newvalue, unitsmoved)


def fitinfield(value, fieldsize):
    "Attempt to fit value into a field, preserving as much data as possible"
    vallen = len(value)
    if fieldsize is None:
        newvalue = value
    elif vallen == fieldsize:  # Goldilocks!
        newvalue = value
    elif vallen < fieldsize:  # Extra room, pad it out
        pad = " " * (fieldsize - vallen)
        newvalue = pad + value
    else:  # Insufficient room, round as few digits as possible
        if "." in value:  # Ok, we *do* have decimals to crop
            diff = vallen - fieldsize
            declen = len(value.split(".")[1])  # length of decimals
            croplen = min(declen, diff)  # Never round above the decimal point
            roundlen = declen - croplen  # How many digits we round to
            newvalue = str(round(float(value), roundlen))
            splitted = newvalue.split(".")  # This should never fail
            declen = len(splitted[1])
            if roundlen == 0:  # if rounding all the decimals don't display .0
                # but do display the point, to show that there is more beyond
                newvalue = splitted[0] + "."
            elif roundlen > declen:  # some zeros have been cropped, fix that
                padcount = roundlen - declen
                newvalue = newvalue + ("0" * padcount)
        else:  # No decimals, nothing we can crop
            newvalue = value
    return newvalue


def cropprecision(value, ooms):
    "Crops digits below the maximum precision"
    if "." not in value:  # No decimals, nothing to crop
        return value
    if ooms == 0:  # We are at the baseunit, crop it all
        return value.split(".")[0]
    dstart = value.find(".") + 1
    dsize = len(value) - dstart
    precision = min(ooms, dsize)
    cropcount = dsize - precision
    if cropcount > 0:
        value = value[:-cropcount]
    return value


def isstringzero(value):
    "Detects whether a string is equal to zero"
    for i in value:
        if i not in ("-", ".", "0"):
            return False
    return True


def unitrelativeto(unit, move):
    "Returns a unit at a different scale from the input unit"
    for group in unitgroups:
        if unit in group:
            if move is None:  # asking for the base unit
                return group[0]
            else:
                index = group.index(unit)
                index += move  # index of the new unit
                if 0 <= index < len(group):  # found the new unit
                    return group[index]
                else:  # not in range
                    return None
    return None  # couldn't find anything


def unitifyvar(value, varname, baseunit=None, width=8, unitSpace=False):
    "Call unitify() with the correct units for varname"
    if varname in S_VARS:
        start = UNIT_S
    elif varname in MS_VARS:
        start = UNIT_MS
    elif varname in PPM_VARS:
        start = UNIT_PPM
    else:
        return value
    return unitify(value, start, baseunit, width, unitSpace)


def unitify(value, startingunit, baseunit=None, width=8, unitSpace=False):
    "Formats a numberstring with relevant units. Attempts to fit in width."
    if baseunit is None:
        baseunit = getunitgroup(startingunit)[0]
    ooms = oomsbetweenunits(startingunit, baseunit)
    if isstringzero(value):
        newvalue, unitsmoved = formatzero(value)
    else:
        newvalue = cropprecision(value, ooms)
        newvalue, unitsmoved = scalestring(newvalue)
    unitget = unitrelativeto(startingunit, unitsmoved)
    if unitSpace:
        spaceWidthAdjustment = 1
        spacer = " "
    else:
        spaceWidthAdjustment = 0
        spacer = ""
    if unitget is not None:  # We have a unit
        if width is None:
            realwidth = None
        else:
            realwidth = width - (len(unitget) + spaceWidthAdjustment)
        newvalue = fitinfield(newvalue, realwidth) + spacer + unitget
    else:  # don't have a replacement unit, use original
        newvalue = value + spacer + startingunit
    if width is None:
        newvalue = newvalue.strip()
    return newvalue


def f8dot4(f):
    "Scaled floating point formatting to fit in 8 characters"

    if isinstance(f, str):
        # a string? pass it on as a signal
        return "%8s" % f
    if not isinstance(f, (int, float)):
        # huh?
        return "       X"
    if str(float(f)).lower() == 'nan':
        # yes, this is a better test than math.isnan()
        # it also catches None, strings, etc.
        return "     nan"

    fmt = "%8d"          # xxxxxxxx or -xxxxxxx
    if f >= 0:
        if f < 1000.0:
            fmt = "%8.4f"    # xxx.xxxx  normal case
        elif f < 10000.0:
            fmt = "%8.3f"    # xxxx.xxx
        elif f < 100000.0:
            fmt = "%8.2f"    # xxxxx.xx
        elif f < 1000000.0:
            fmt = "%8.1f"    # xxxxxx.x
    else:
        # negative number, account for minus sign
        if f > -100.0:
            fmt = "%8.4f"      # -xx.xxxx  normal case
        elif f > -1000.0:
            fmt = "%8.3f"      # -xxx.xxx
        elif f > -10000.0:
            fmt = "%8.2f"      # -xxxx.xx
        elif f > -100000.0:
            fmt = "%8.1f"      # -xxxxx.x

    return fmt % f


def f8dot3(f):
    "Scaled floating point formatting to fit in 8 characters"
    if isinstance(f, str):
        # a string? pass it on as a signal
        return "%8s" % f
    if not isinstance(f, (int, float)):
        # huh?
        return "       X"
    if str(float(f)).lower() == 'nan':
        # yes, this is a better test than math.isnan()
        # it also catches None, strings, etc.
        return "     nan"

    fmt = "%8d"          # xxxxxxxx or -xxxxxxx
    if f >= 0:
        if f < 10000.0:
            fmt = "%8.3f"    # xxxx.xxx  normal case
        elif f < 100000.0:
            fmt = "%8.2f"    # xxxxx.xx
        elif f < 1000000.0:
            fmt = "%8.1f"    # xxxxxx.x
    else:
        # negative number, account for minus sign
        if f > -1000.0:
            fmt = "%8.3f"    # -xxx.xxx  normal case
        elif f > -10000.0:
            fmt = "%8.2f"    # -xxxx.xx
        elif f > -100000.0:
            fmt = "%8.1f"    # -xxxxx.x

    return fmt % f


def monoclock():
    "Try to get a monotonic clock value unaffected by NTP stepping."
    try:
        # Available in Python 3.3 and up.
        return time.monotonic()
    except AttributeError:
        return time.time()


class Cache:
    "Simple time-based cache"

    def __init__(self, defaultTimeout=300):  # 5 min default TTL
        self.defaultTimeout = defaultTimeout
        self._cache = {}

    def get(self, key):
        if key in self._cache:
            value, settime, ttl = self._cache[key]
            if settime >= monoclock() - ttl:
                return value
            else:  # key expired, delete it
                del self._cache[key]
                return None
        else:
            return None

    def set(self, key, value, customTTL=None):
        ttl = customTTL if customTTL is not None else self.defaultTimeout
        self._cache[key] = (value, monoclock(), ttl)


# A hack to avoid repeatedly hammering on DNS when ntpmon runs.
canonicalization_cache = Cache()


def canonicalize_dns(inhost, family=socket.AF_UNSPEC):
    "Canonicalize a hostname or numeric IP address."
    resname = canonicalization_cache.get(inhost)
    if resname is not None:
        return resname
    # Catch garbaged hostnames in corrupted Mode 6 responses
    m = re.match(r"([:.[\]]|\w)*", inhost)
    if not m:
        raise TypeError
    (hostname, portsuffix) = portsplit(inhost)
    try:
        ai = socket.getaddrinfo(hostname, None, family, 0, 0,
                                socket.AI_CANONNAME)
    except socket.gaierror:
        return "DNSFAIL:%s" % hostname
    (family, socktype, proto, canonname, sockaddr) = ai[0]
    try:
        name = socket.getnameinfo(sockaddr, socket.NI_NAMEREQD)
        result = name[0].lower() + portsuffix
    except socket.gaierror:
        # On OS X, canonname is empty for hosts without rDNS.
        # Fall back to the hostname.
        canonicalized = canonname or hostname
        result = canonicalized.lower() + portsuffix
    canonicalization_cache.set(inhost, result)
    return result


TermSize = collections.namedtuple("TermSize", ["width", "height"])


# Python 2.x does not have the shutil.get_terminal_size function.
# This conditional import is only needed by termsize() and should be kept
# near it. It is not inside the function because the unit tests need to be
# able to splice in a jig.
if str is bytes:  # We are on python 2.x
    import fcntl
    import termios
    import struct


def termsize():  # pragma: no cover
    "Return the current terminal size."
    # Alternatives at http://stackoverflow.com/questions/566746
    # The way this is used makes it not a big deal if the default is wrong.
    size = (80, 24)
    if os.isatty(1):
        if str is not bytes:
            # str is bytes means we are >py3.0, but this will still fail
            # on versions <3.3. We do not support those anyway.
            (w, h) = shutil.get_terminal_size((80, 24))
            size = (w, h)
        else:
            try:
                # OK, Python version < 3.3, cope
                h, w, hp, wp = struct.unpack(
                    'HHHH',
                    fcntl.ioctl(2, termios.TIOCGWINSZ,
                                struct.pack('HHHH', 0, 0, 0, 0)))
                size = (w, h)
            except IOError:
                pass
    return TermSize(*size)


class PeerStatusWord:
    "A peer status word from readstats(), dissected for display"

    def __init__(self, status, pktversion=ntp.magic.NTP_VERSION):
        # Event
        self.event = ntp.control.CTL_PEER_EVENT(status)
        # Event count
        self.event_count = ntp.control.CTL_PEER_NEVNT(status)
        statval = ntp.control.CTL_PEER_STATVAL(status)
        # Config
        if statval & ntp.control.CTL_PST_CONFIG:
            self.conf = "yes"
        else:
            self.conf = "no"
        # Reach
        if statval & ntp.control.CTL_PST_BCASTx:
            self.reach = "none"
        elif statval & ntp.control.CTL_PST_REACH:
            self.reach = "yes"
        else:
            self.reach = "no"
        # Auth
        if (statval & ntp.control.CTL_PST_AUTHENABLE) == 0:
            self.auth = "none"
        elif statval & ntp.control.CTL_PST_AUTHENTIC:
            self.auth = "ok "
        else:
            self.auth = "bad"
        # Condition
        if pktversion > ntp.magic.NTP_OLDVERSION:
            seldict = {
                ntp.control.CTL_PST_SEL_REJECT: "reject",
                ntp.control.CTL_PST_SEL_SANE: "falsetick",
                ntp.control.CTL_PST_SEL_CORRECT: "excess",
                ntp.control.CTL_PST_SEL_SELCAND: "outlier",
                ntp.control.CTL_PST_SEL_SYNCCAND: "candidate",
                ntp.control.CTL_PST_SEL_EXCESS: "backup",
                ntp.control.CTL_PST_SEL_SYSPEER: "sys.peer",
                ntp.control.CTL_PST_SEL_PPS: "pps.peer",
                }
            self.condition = seldict[statval & 0x7]
        else:
            if (statval & 0x3) == OLD_CTL_PST_SEL_REJECT:
                if (statval & OLD_CTL_PST_SANE) == 0:
                    self.condition = "insane"
                elif (statval & OLD_CTL_PST_DISP) == 0:
                    self.condition = "hi_disp"
                else:
                    self.condition = ""
            elif (statval & 0x3) == OLD_CTL_PST_SEL_SELCAND:
                self.condition = "sel_cand"
            elif (statval & 0x3) == OLD_CTL_PST_SEL_SYNCCAND:
                self.condition = "sync_cand"
            elif (statval & 0x3) == OLD_CTL_PST_SEL_SYSPEER:
                self.condition = "sys_peer"
        # Last Event
        event_dict = {
            ntp.magic.PEVNT_MOBIL: "mobilize",
            ntp.magic.PEVNT_DEMOBIL: "demobilize",
            ntp.magic.PEVNT_UNREACH: "unreachable",
            ntp.magic.PEVNT_REACH: "reachable",
            ntp.magic.PEVNT_RESTART: "restart",
            ntp.magic.PEVNT_REPLY: "no_reply",
            ntp.magic.PEVNT_RATE: "rate_exceeded",
            ntp.magic.PEVNT_DENY: "access_denied",
            ntp.magic.PEVNT_ARMED: "leap_armed",
            ntp.magic.PEVNT_NEWPEER: "sys_peer",
            ntp.magic.PEVNT_CLOCK: "clock_alarm",
            }
        self.last_event = event_dict.get(ntp.magic.PEER_EVENT | self.event, "")

    def __str__(self):
        return ("conf=%(conf)s, reach=%(reach)s, auth=%(auth)s, "
                "cond=%(condition)s, event=%(last_event)s ec=%(event_count)s"
                % self.__dict__)


def cook(variables, showunits=False, sep=", "):
    "Cooked-mode variable display."
    width = ntp.util.termsize().width - 2
    text = ""
    specials = ("filtdelay", "filtoffset", "filtdisp", "filterror")
    longestspecial = len(max(specials, key=len))
    for (name, (value, rawvalue)) in variables.items():
        if name in specials:  # need special formatting for column alignment
            formatter = "%" + str(longestspecial) + "s ="
            item = formatter % name
        else:
            item = "%s=" % name
        if name in ("reftime", "clock", "org", "rec", "xmt"):
            item += ntp.ntpc.prettydate(value)
        elif name in ("srcadr", "peeradr", "dstadr", "refid"):
            # C ntpq cooked these in obscure ways.  Since they
            # came up from the daemon as human-readable
            # strings this was probably a bad idea, but we'll
            # leave this case separated in case somebody thinks
            # re-cooking them is a good idea.
            item += value
        elif name == "leap":
            item += ("00", "01", "10", "11")[value]
        elif name == "reach":
            item += "%03lo" % value
        elif name in specials:
            if showunits:
                item += stringfiltcooker(value)
            else:
                item += "\t".join(value.split())
        elif name == "flash":
            item += "%02x " % value
            if value == 0:
                item += "ok "
            else:
                # flasher bits
                tstflagnames = (
                    "pkt_dup",          # BOGON1
                    "pkt_bogus",        # BOGON2
                    "pkt_unsync",       # BOGON3
                    "pkt_denied",       # BOGON4
                    "pkt_auth",         # BOGON5
                    "pkt_stratum",      # BOGON6
                    "pkt_header",       # BOGON7
                    "pkt_autokey",      # BOGON8
                    "pkt_crypto",       # BOGON9
                    "peer_stratum",     # BOGON10
                    "peer_dist",        # BOGON11
                    "peer_loop",        # BOGON12
                    "peer_unreach"      # BOGON13
                )
                for (i, n) in enumerate(tstflagnames):
                    if (1 << i) & value:
                        item += tstflagnames[i] + " "
            item = item[:-1]
        elif name in MS_VARS:
            #  Note that this is *not* complete, there are definitely
            #   missing variables here.
            #  Completion cannot occur until all units are tracked down.
            if showunits:
                item += unitify(rawvalue, UNIT_MS, UNIT_NS, width=None)
            else:
                item += repr(value)
        elif name in S_VARS:
            if showunits:
                item += unitify(rawvalue, UNIT_S, UNIT_NS, width=None)
            else:
                item += repr(value)
        elif name in PPM_VARS:
            if showunits:
                item += unitify(rawvalue, UNIT_PPM, width=None)
            else:
                item += repr(value)
        else:
            item += repr(value)
        # add field separator
        item += sep
        # add newline so we don't overflow screen
        lastcount = 0
        for c in text:
            if c == '\n':
                lastcount = 0
            else:
                lastcount += 1
        if lastcount + len(item) > width:
            text = text[:-1] + "\n"
        text += item
    text = text[:-2] + "\n"
    return text


class PeerSummary:
    "Reusable report generator for peer statistics"

    def __init__(self, displaymode, pktversion, showhostnames,
                 wideremote, showunits=False, termwidth=None,
                 debug=0, logfp=sys.stderr):
        self.displaymode = displaymode          # peers/apeers/opeers
        self.pktversion = pktversion            # interpretation of flash bits
        self.showhostnames = showhostnames      # If false, display numeric IPs
        self.showunits = showunits              # If False show old style float
        self.wideremote = wideremote            # show wide remote names?
        self.debug = debug
        self.logfp = logfp
        self.termwidth = termwidth
        # By default, the peer spreadsheet layout is designed so lines just
        # fit in 80 characters. This tells us how much extra horizontal space
        # we have available on a wider terminal emulator.
        self.horizontal_slack = min((termwidth or 80) - 80, 24)
        # Peer spreadsheet column widths. The reason we cap extra
        # width used at 24 is that on very wide displays, slamming the
        # non-hostname fields all the way to the right produces a huge
        # river that makes the entries difficult to read as wholes.
        # This choice caps the peername field width at that of the longest
        # possible IPV6 numeric address.
        self.namewidth = 15 + self.horizontal_slack
        self.refidwidth = 15
        # Compute peer spreadsheet headers
        self.__remote = "     remote    ".ljust(self.namewidth)
        self.__common = "st t when poll reach   delay   offset   "
        self.__header = None
        self.polls = []

    @staticmethod
    def prettyinterval(diff):
        "Print an interval in natural time units."
        if not isinstance(diff, int) or diff <= 0:
            return '-'
        if diff <= 2048:
            return str(diff)
        diff = (diff + 29) / 60
        if diff <= 300:
            return "%dm" % diff
        diff = (diff + 29) / 60
        if diff <= 96:
            return "%dh" % diff
        diff = (diff + 11) / 24
        return "%dd" % diff

    @staticmethod
    def high_truncate(hostname, maxlen):
        "Truncate on the left using leading _ to indicate 'more'."
        # Used for local IPv6 addresses, best distinguished by low bits
        if len(hostname) <= maxlen:
            return hostname
        else:
            return '-' + hostname[-maxlen+1:]

    @staticmethod
    def is_clock(variables):
        "Does a set of variables look like it returned from a clock?"
        return "srchost" in variables and '(' in variables["srchost"][0]

    def header(self):
        "Column headers for peer display"
        if self.displaymode == "apeers":
            self.__header = self.__remote + \
                "   refid   assid  ".ljust(self.refidwidth) + \
                self.__common + "jitter"
        elif self.displaymode == "opeers":
            self.__header = self.__remote + \
                "       local      ".ljust(self.refidwidth) + \
                self.__common + "  disp"
        elif self.displaymode == 'rpeers':
            self.__header = ' st t when poll reach   delay   ' + \
                            'offset   jitter refid           T remote'
        else:
            self.__header = self.__remote + \
                "       refid      ".ljust(self.refidwidth) + \
                self.__common + "jitter"
        return self.__header

    def width(self):
        "Width of display"
        return 79 + self.horizontal_slack

    def summary(self, rstatus, variables, associd):
        "Peer status summary line."
        clock_name = ''
        dstadr_refid = ""
        dstport = 0
        estdelay = '.'
        estdisp = '.'
        estjitter = '.'
        estoffset = '.'
        filtdelay = 0.0
        filtdisp = 0.0
        filtoffset = 0.0
        flash = 0
        have_jitter = False
        headway = 0
        hmode = 0
        hpoll = 0
        keyid = 0
        last_sync = None
        leap = 0
        pmode = 0
        ppoll = 0
        precision = 0
        ptype = '?'
        reach = 0
        rec = None
        reftime = None
        rootdelay = 0.0
        saw6 = False        # x.6 floats for delay and friends
        srcadr = None
        srchost = None
        srcport = 0
        stratum = 20
        mode = 0
        unreach = 0
        xmt = 0
        ntscookies = -1

        now = time.time()

        for item in variables.items():
            if 2 != len(item) or 2 != len(item[1]):
                # bad item
                continue
            (name, (value, rawvalue)) = item
            if name == "delay":
                estdelay = rawvalue if self.showunits else value
                if len(rawvalue) > 6 and rawvalue[-7] == ".":
                    saw6 = True
            elif name == "dstadr":
                # The C code tried to get a fallback ptype from this in case
                # the hmode field was not included
                if "local" in self.__header:
                    dstadr_refid = rawvalue
            elif name == "dstport":
                # FIXME, dstport never used.
                dstport = value
            elif name == "filtdelay":
                # FIXME, filtdelay never used.
                filtdelay = value
            elif name == "filtdisp":
                # FIXME, filtdisp never used.
                filtdisp = value
            elif name == "filtoffset":
                # FIXME, filtoffset never used.
                filtoffset = value
            elif name == "flash":
                # FIXME, flash never used.
                flash = value
            elif name == "headway":
                # FIXME, headway never used.
                headway = value
            elif name == "hmode":
                hmode = value
            elif name == "hpoll":
                hpoll = value
                if hpoll < 0:
                    hpoll = ntp.magic.NTP_MINPOLL
            elif name == "jitter":
                if "jitter" in self.__header:
                    estjitter = rawvalue if self.showunits else value
                    have_jitter = True
            elif name == "keyid":
                # FIXME, keyid never used.
                keyid = value
            elif name == "leap":
                # FIXME, leap never used.
                leap = value
            elif name == "offset":
                estoffset = rawvalue if self.showunits else value
            elif name == "pmode":
                # FIXME, pmode never used.
                pmode = value
            elif name == "ppoll":
                ppoll = value
                if ppoll < 0:
                    ppoll = ntp.magic.NTP_MINPOLL
            elif name == "precision":
                # FIXME, precision never used.
                precision = value
            elif name == "reach":
                # Shipped as hex, displayed in octal
                reach = value
            elif name == "refid":
                # The C code for this looked crazily overelaborate.  Best
                # guess is that it was designed to deal with formats that
                # no longer occur in this field.
                if "refid" in self.__header:
                    dstadr_refid = rawvalue
            elif name == "rec":
                rec = value         # l_fp timestamp
                last_sync = int(now - ntp.ntpc.lfptofloat(rec))
            elif name == "reftime":
                reftime = value     # l_fp timestamp
                last_sync = int(now - ntp.ntpc.lfptofloat(reftime))
            elif name == "rootdelay":
                # FIXME, rootdelay never used.
                rootdelay = value   # l_fp timestamp
            elif name == "rootdisp" or name == "dispersion":
                estdisp = rawvalue if self.showunits else value
            elif name in ("srcadr", "peeradr"):
                srcadr = value
            elif name == "srchost":
                srchost = value
            elif name == "srcport" or name == "peerport":
                # FIXME, srcport never used.
                srcport = value
            elif name == "stratum":
                stratum = value
            elif name == "mode":
                # FIXME, mode never used.
                mode = value
            elif name == "unreach":
                # FIXME, unreach never used.
                unreach = value
            elif name == "xmt":
                # FIXME, xmt never used.
                xmt = value
            elif name == "ntscookies":
                ntscookies = value
            else:
                # unknown name?
                # line = " name=%s " % (name)    # debug
                # return line                    # debug
                continue
        if hmode == ntp.magic.MODE_BCLIENTX:
            # broadcastclient or multicastclient
            ptype = 'b'
        elif hmode == ntp.magic.MODE_BROADCASTx:
            # broadcast or multicast server
            if srcadr.startswith("224."):       # IANA multicast address prefix
                ptype = 'M'
            else:
                ptype = 'B'
        elif hmode == ntp.magic.MODE_CLIENT:
            if PeerSummary.is_clock(variables):
                ptype = 'l'     # local refclock
            elif dstadr_refid == "POOL":
                ptype = 'p'     # pool
            elif srcadr.startswith("224."):
                ptype = 'a'     # manycastclient (compatibility with Classic)
            elif ntscookies > -1:
                # FIXME: Will foo up if there are ever more than 9 cookies
                ptype = chr(ntscookies + ord('0'))
            else:
                ptype = 'u'     # unicast
        elif hmode == ntp.magic.MODE_ACTIVEx:
            ptype = 's'         # symmetric active
        elif hmode == ntp.magic.MODE_PASSIVEx:
            ptype = 'S'         # symmetric passive

        #
        # Got everything, format the line
        #
        line = ""
        poll_sec = 1 << min(ppoll, hpoll)
        self.polls.append(poll_sec)
        if self.pktversion > ntp.magic.NTP_OLDVERSION:
            c = " x.-+#*o"[ntp.control.CTL_PEER_STATVAL(rstatus) & 0x7]
        else:
            c = " .+*"[ntp.control.CTL_PEER_STATVAL(rstatus) & 0x3]
        # Source host or clockname or poolname or servername
        # After new DNS, 2017-Apr-17
        # servers setup via numerical IP Address have only srcadr
        # servers setup via DNS have both srcadr and srchost
        # refclocks have both srcadr and srchost
        # pool has "0.0.0.0" (or "::") and srchost
        # slots setup via pool have only srcadr
        if srcadr is not None \
                and srcadr != "0.0.0.0" \
                and not srcadr.startswith("127.127") \
                and srcadr != "::":
            if self.showhostnames & 2 and 'srchost' in locals() and srchost:
                clock_name = srchost
            elif self.showhostnames & 1:
                try:
                    if self.debug:
                        self.logfp.write("DNS lookup begins...\n")
                    clock_name = canonicalize_dns(srcadr)
                    if self.debug:
                        self.logfp.write("DNS lookup ends.\n")
                except TypeError:  # pragma: no cover
                    return ''
            else:
                clock_name = srcadr
        else:
            clock_name = srchost
        if clock_name is None:
            if srcadr:
                clock_name = srcadr
            else:
                clock_name = ""
        if self.displaymode != "rpeers":
            if self.wideremote and len(clock_name) > self.namewidth:
                line += ("%c%s\n" % (c, clock_name))
                line += (" " * (self.namewidth + 2))
            else:
                line += ("%c%-*.*s " % (c, self.namewidth, self.namewidth,
                                        clock_name[:self.namewidth]))
        # Destination address, assoc ID or refid.
        assocwidth = 7 if self.displaymode == "apeers" else 0
        if "." not in dstadr_refid and ":" not in dstadr_refid:
            dstadr_refid = "." + dstadr_refid + "."
        if assocwidth and len(dstadr_refid) >= self.refidwidth - assocwidth:
            visible = "..."
        else:
            visible = dstadr_refid
        if self.displaymode != "rpeers":
            line += self.high_truncate(visible, self.refidwidth)
            if self.displaymode == "apeers":
                line += (" " * (self.refidwidth - len(visible) - assocwidth + 1))
                line += ("%-6d" % (associd))
            else:
                line += (" " * (self.refidwidth - len(visible)))
        # The rest of the story
        if last_sync is None:
            last_sync = now
        jd = estjitter if have_jitter else estdisp
        if self.showunits:
            fini = lambda x : unitify(x, UNIT_MS)
        elif saw6:
            fini = lambda x : f8dot4(x)
        else:
            fini = lambda x : f8dot3(x)
        line += (
            " %2ld %c %4.4s %4.4s  %3lo %s %s %s"
            % (stratum, ptype,
               PeerSummary.prettyinterval(last_sync),
               PeerSummary.prettyinterval(poll_sec), reach,
               fini(estdelay), fini(estoffset), fini(jd)))
        if self.displaymode == "rpeers":
            line += " " + dstadr_refid
            line += (" " * (self.refidwidth - len(dstadr_refid)))
            line += (" %c %s" % (c, clock_name[:76 - len(line)]))
        line += "\n"
        # for debugging both case
        # if srcadr != None and srchost != None:
        #   line += "srcadr: %s, srchost: %s\n" % (srcadr, srchost)
        return line

    def intervals(self):
        "Return and flush the list of actual poll intervals."
        res = self.polls[:]
        self.polls = []
        return res


class MRUSummary:
    "Reusable class for MRU entry summary generation."

    def __init__(self, showhostnames, wideremote=False,
                 debug=0, logfp=sys.stderr):
        self.debug = debug
        self.logfp = logfp
        self.now = None
        self.showhostnames = showhostnames  # if & 1, display names
        self.wideremote = wideremote

    header = " lstint avgint rstr r m v  count    score   drop rport remote address"

    def summary(self, entry):
        first = ntp.ntpc.lfptofloat(entry.first)
        last = ntp.ntpc.lfptofloat(entry.last)
        active = float(last - first)
        count = int(entry.ct)
        if self.now:
            lstint = int(self.now - last + 0.5)
            stats = "%7d" % lstint
            if count == 1:
                favgint = 0
            else:
                favgint = active / (count-1)
            avgint = int(favgint + 0.5)
            if 5.0 < favgint or 1 == count:
                stats += " %6d" % avgint
            elif 1.0 <= favgint:
                stats += " %6.2f" % favgint
            else:
                stats += " %6.3f" % favgint
        else:
            MJD_1970 = 40587     # MJD for 1 Jan 1970, Unix epoch
            days, lstint = divmod(int(last), 86400)
            stats = "%5d %5d %6d" % (days + MJD_1970, lstint, active)
        if entry.rs & ntp.magic.RES_KOD:
            rscode = 'K'
        elif entry.rs & ntp.magic.RES_LIMITED:
            rscode = 'L'
        else:
            rscode = '.'
        (ip, port) = portsplit(entry.addr)
        try:
            if not self.showhostnames & 1:  # if not & 1 display numeric IPs
                dns = ip
            else:
                dns = canonicalize_dns(ip)
                # Forward-confirm the returned DNS
                confirmed = canonicalization_cache.get(dns)
                if confirmed is None:
                    confirmed = False
                    try:
                        ai = socket.getaddrinfo(dns, None)
                        for (_, _, _, _, sockaddr) in ai:
                            if sockaddr and sockaddr[0] == ip:
                                confirmed = True
                                break
                    except socket.gaierror:
                        pass
                    canonicalization_cache.set(dns, confirmed)
                if not confirmed:
                    dns = "%s (%s)" % (ip, dns)
            if not self.wideremote:
                # truncate for narrow display
                dns = dns[:40]
            if entry.sc:
                score = float(entry.sc)
                if score > 100000.0:
                  score = "%8.1f" % score
                elif score > 10000.0:
                  score = "%8.2f" % score
                else:
                  score = "%8.3f" % score
            else:
                score = "-"
            if entry.dr!= None:     # 0 is valid
                drop = "%4d" % entry.dr
            else:
                drop = "-"
            stats += " %4hx %c %d %d %6d %8s %6s %5s %s" % \
                     (entry.rs, rscode,
                      ntp.magic.PKT_MODE(entry.mv),
                      ntp.magic.PKT_VERSION(entry.mv),
                      entry.ct, score, drop, port[1:], dns)
            return stats
        except ValueError:
            # This can happen when ntpd ships a corrupt varlist
            return ''


class ReslistSummary:
    "Reusable class for reslist entry summary generation."
    header = """\
   hits    addr/prefix or addr mask
           restrictions
"""
    width = 72

    @staticmethod
    def __getPrefix(mask):
        if not mask:
            prefix = ''
        if ':' in mask:
            sep = ':'
            base = 16
        else:
            sep = '.'
            base = 10
        prefix = sum([bin(int(x, base)).count('1')
                      for x in mask.split(sep) if x])
        return '/' + str(prefix)

    def summary(self, variables):
        hits = variables.get("hits", "?")
        address = variables.get("addr", "?")
        mask = variables.get("mask", "?")
        if address == '?' or mask == '?':
            return ''
        address += ReslistSummary.__getPrefix(mask)
        flags = variables.get("flags", "?")
        # reslist responses are often corrupted
        s = "%10s %s\n           %s\n" % (hits, address, flags)
        # Throw away corrupted entries.  This is a shim - we really
        # want to make ntpd stop generating garbage
        for c in s:
            if not c.isalnum() and c not in "/.: \n":
                return ''
        return s


class IfstatsSummary:
    "Reusable class for ifstats entry summary generation."
    header = """\
    interface name                                  send
 #  address/broadcast     drop flag received sent failed peers   uptime
 """
    width = 74
    # Numbers are the fieldsize
    fields = {'name':  '%-24.24s',
              'flags': '%4x',
              'rx':    '%6d',
              'tx':    '%6d',
              'txerr': '%6d',
              'pc':    '%5d',
              'up':    '%8d'}

    def summary(self, i, variables):
        formatted = {}
        try:
            # Format the fields
            for name in self.fields.keys():
                value = variables.get(name, "?")
                if value == "?":
                    fmt = value
                else:
                    fmt = self.fields[name] % value
                formatted[name] = fmt
            enFlag = '.' if variables.get('en', False) else 'D'
            address = variables.get("addr", "?")
            bcast = variables.get("bcast")
            # Assemble the fields into a line
            s = ("%3u %s %s %s %s %s %s %s %s\n    %s\n"
                 % (i,
                    formatted['name'],
                    enFlag,
                    formatted['flags'],
                    formatted['rx'],
                    formatted['tx'],
                    formatted['txerr'],
                    formatted['pc'],
                    formatted['up'],
                    address))
            if bcast:
                s += "    %s\n" % bcast
        except TypeError:  # pragma: no cover
            # Can happen when ntpd ships a corrupted response
            return ''

        # FIXME, a brutal and slow way to check for invalid chars..
        # maybe just strip non-printing chars?
        for c in s:
            if not c.isalnum() and c not in "/.:[] %\n":
                return ''
        return s


def packetize(packets, period, clipdigits=0, periodized=False):
    """Given a number of packets and a duration (s) return a tuple.

    return the packet quantity, and a two part rate in packets/seconds
    or seconds/packet. On error the latter fields should be blank, the
    first the number of packets if zero otherwise unhelpful text."""
    if not isinstance(packets, int):
        return ("???", "", "")
    if packets == 0 or not isinstance(period, (int, float)):
        return (packets, "", "")
    if packets > period:
        return (packets, round(packets / period, clipdigits), "p/s")
    if periodized:
        return (packets, periodize(period / packets, clipdigits)[1], "/p")
    return (packets, round(period / packets, clipdigits), "s/p")


def periodize(period, clipdigits=0):
    """Given a number of seconds, return number and pretty string.

    On error return None for the number and an unhelpful string."""
    clip = clipdigits if isinstance(clipdigits, (int, float)) else 0
    if not isinstance(period, (int, float)):
        return (None, "???")
    result = ""
    _ = round(period, clip)
    nperiod = int(_) if clip < 1 else _
    if nperiod >= 86400:
        result += "%dD " % (nperiod // 86400)
    result += "%02d:%02d:%02d" % (
        (nperiod % 86400) // 3600,
        (nperiod % 3600) // 60,
        nperiod % 60,
    )
    return (nperiod, result)


uptime = lambda p: periodize(p)[1]
# end



# Copyright the NTPsec project contributors
#
# SPDX-License-Identifier: BSD-2-Clause

def options_cmd(ctx, config):
    ctx.load("compiler_c")
    ctx.load("msvc")
    ctx.load('waf_unit_test')
    ctx.load('gnu_dirs')

    grp = ctx.add_option_group("NTP configure options")
    grp.add_option('--enable-debug', action='store_true',
                   default=False, help="Enable debugging code")
    grp.add_option('--disable-debug-gdb', action='store_true',
                   default=False, help="Disable GDB debugging symbols")
    grp.add_option('--enable-attic', action='store_true',
                   default=False, help="Enable building attic/*.")
    grp.add_option('--disable-nts', action='store_true',
                   default=False, help="Disable NTS.")
    grp.add_option('--disable-droproot', action='store_true',
                   default=False, help="Disable dropping root.")
    grp.add_option('--enable-early-droproot', action='store_true',
                   default=False,
                   help="Droproot earlier (breaks SHM and NetBSD).")
    grp.add_option('--enable-seccomp', action='store_true',
                   default=False, help="Enable seccomp (restricts syscalls).")
    grp.add_option('--disable-mdns-registration', action='store_true',
                   default=False, help="Disable MDNS registration.")
    grp.add_option(
        '--enable-classic-mode', action='store_true',
        default=False,
        help="Strict configuration and log-format "
             " compatibility with NTP Classic")
    grp.add_option('--enable-debug-timing', action='store_true',
                   default=False,
                   help="Collect timing statistics for debugging.")
    grp.add_option('--enable-pylib', action='store',
                   default='ffi', choices=['ext', 'ffi', 'none'],
                   help="""Choose which Python library to build.\n
ext, ffi, or none. defaults to ffi.""", nargs=1)

    grp = ctx.add_option_group("NTP cross compile options")
    grp.add_option('--cross-compiler', type=str,
                   help="Path to cross compiler CC. (enables cross-compiling)")
    grp.add_option('--cross-cflags', type=str, action='append',
                   help="Cross compiler CFLAGS.")
    grp.add_option('--cross-ldflags', type=str, action='append',
                   help="Cross compiler LDFLAGS.")

    grp = ctx.add_option_group("NTP configure features")
    grp.add_option('--enable-leap-smear', action='store_true',
                   default=False, help="Enable Leap Smearing.")
    grp.add_option('--enable-leap-testing', action='store_true',
                   default=False,
                   help="Enable leaps on other than 1st of month.")
    grp.add_option('--enable-mssntp', action='store_true',
                   default=False, help="Enable Samba MS SNTP support.")

    grp = ctx.add_option_group("Refclock configure options")
    grp.add_option(
        '--refclock', dest='refclocks',
        help="Comma-separated list of Refclock IDs to build (or \"all\")",
        type=str)
    grp.add_option('--list', action='store_true', default=False,
                   help="List available Refclocks")

    grp = ctx.add_option_group("NTP developer configure options")
    grp.add_option('--build-version-tag', type=str,
                   help="Append a tag to the version string (unused)")
    grp.add_option('--cflags', type=str, action='append',
                   help="Users should use CFLAGS in their environment.")
    grp.add_option('--build-desc', type=str, default="",
                   help="Arbitrary text string to append to extended version.")
    grp.add_option('--ldflags', type=str, action='append',
                   help="Users should use LDFLAGS in their environment.")
    grp.add_option('--check', action='store_true', default=False,
                   help="Run tests")
    grp.add_option('--enable-warnings', action='store_true',
                   default=False, help="Turn on a few more warnings")
    grp.add_option('--enable-Werror', action='store_true',
                   default=False, help="Turn on -Werror")
    grp.add_option(
        '--define', type=str, action='append',
        help="Force definition of symbol, wants value of form SYM=VAL.")
    grp.add_option('--undefine', type=str, action='append',
                   help="Force undefinition of symbol.")
    grp.add_option('--disable-fuzz', action='store_true',
                   default=False,
                   help="Feature gone")

    # This option group name needs to exactly match the name from
    # waflib/Tools/python.py so this option groups with the other Python
    # options.
    grp = ctx.add_option_group("Python Options")
    grp.add_option('--pyshebang', type=str,
                   default='/usr/bin/env python',
                   help="Shebang for Python scripts [Default: %(default)s}")



# -*- coding: utf-8 -*-

# Copyright the NTPsec project contributors
#
# SPDX-License-Identifier: BSD-2-Clause

from __future__ import print_function, division

import socket
import select
import os.path
import time
import ntp.poly


class FileJig:
    def __init__(self, returns=[""]):
        self.data = []
        self.flushed = False
        self.readline_return = returns

    def __enter__(self):
        return self

    def __exit__(self, *args):
        return False

    def __iter__(self):
        return self.readline_return.__iter__()

    def write(self, data):
        self.data.append(data)
        self.flushed = False

    def flush(self):
        self.flushed = True

    def readline(self):
        if self.readline_return:
            return self.readline_return.pop(0)
        return ""

    def readlines(self):
        ret = self.readline_return
        self.readline_return = []
        return ret


class SocketJig:
    def __init__(self):
        self.data = []
        self.return_data = []
        self.closed = False
        self.connected = None
        self.fail_connect = False
        self.fail_send = 0

    def sendall(self, data):
        if self.fail_send > 0:
            self.fail_send -= 1
            raise socket.error()
        self.data.append(data)

    def close(self):
        self.closed = True

    def connect(self, addr):
        if self.fail_connect:
            err = socket.error()
            err.strerror = "socket!"
            err.errno = 16
            raise err
        self.connected = addr

    def recv(self, bytecount):
        if self.return_data:
            current = self.return_data.pop(0)
            if len(current) > bytecount:
                ret = current[:bytecount]
                current = current[bytecount:]
                self.return_data.insert(0, current)  # push unwanted data
                return ret
            else:
                return current
        return None


class HasherJig:
    def __init__(self):
        self.update_calls = []
        self.digest_size = 0

    def update(self, data):
        self.update_calls.append(data)
        if data:
            self.digest_size += 1

    def digest(self):
        return ntp.poly.polybytes("blah" * 4)  # 16 byte hash


class SocketModuleJig:
    error = socket.error
    gaierror = socket._socket.gaierror
    herror = socket.herror
    SOCK_DGRAM = socket.SOCK_DGRAM
    IPPROTO_UDP = socket.IPPROTO_UDP
    AF_UNSPEC = socket.AF_UNSPEC
    AI_NUMERICHOST = socket.AI_NUMERICHOST
    AI_CANONNAME = socket.AI_CANONNAME
    EAI_NONAME = socket.EAI_NONAME
    if hasattr(socket, "EAI_NODATA"):
        EAI_NODATA = socket.EAI_NODATA
    else:  # FreeBSD does not have EAI_NODATA (removed from POSIX)
        EAI_NODATA = None
    NI_NAMEREQD = socket.NI_NAMEREQD

    def __init__(self):
        self.gai_calls = []
        self.gai_error_count = 0
        self.gai_returns = []
        self.gni_calls = []
        self.gni_error_count = 0
        self.gni_returns = []
        self.socket_calls = []
        self.socket_fail = False
        self.socket_fail_connect = False
        self.socketsReturned = []
        self.inet_ntop_calls = []
        self.getfqdn_calls = []
        self.getfqdn_returns = []
        self.ghba_calls = []
        self.ghba_returns = []

    def getaddrinfo(self, host, port, family=None, socktype=None,
                    proto=None, flags=None):
        self.gai_calls.append((host, port, family, socktype, proto, flags))
        if self.gai_error_count > 0:
            self.gai_error_count -= 1
            err = self.gaierror("blah")
            err.errno = socket.EAI_NONAME
            raise err
        return self.gai_returns.pop(0)

    def getnameinfo(self, addr, flags):
        self.gni_calls.append((addr, flags))
        if self.gni_error_count > 0:
            self.gni_error_count -= 1
            err = self.gaierror("blah")
            err.errno = socket.EAI_NONAME
            raise err
        return self.gni_returns.pop(0)

    def socket(self, family, socktype, protocol):
        self.socket_calls.append((family, socktype, protocol))
        if self.socket_fail:
            err = self.error()
            err.strerror = "error!"
            err.errno = 23
            raise err
        sock = SocketJig()
        if self.socket_fail_connect:
            sock.fail_connect = True
        self.socketsReturned.append(sock)
        return sock

    def inet_ntop(self, addr, family):
        self.inet_ntop_calls.append((addr, family))
        return "canon.com"

    def getfqdn(self, name=""):
        self.getfqdn_calls.append(name)
        return self.getfqdn_returns.pop(0)

    def gethostbyaddr(self, addr):
        self.ghba_calls.append(addr)
        ret = self.ghba_returns.pop(0)
        if ret is None:
            raise self.herror
        return ret


class GetpassModuleJig:
    def __init__(self):
        self.getpass_calls = []
        self.getpass_returns = []

    def getpass(self, prompt, stream=None):
        self.getpass_calls.append((prompt, stream))
        return self.getpass_returns.pop(0)


class HashlibModuleJig:
    def __init__(self):
        self.new_calls = []
        self.hashers_returned = []

    def new(self, name):
        self.new_calls.append(name)
        h = HasherJig()
        self.hashers_returned.append(h)
        return h


class SelectModuleJig:
    error = select.error

    def __init__(self):
        self.select_calls = []
        self.select_fail = 0
        self.do_return = []
        self.fqdn_calls = []
        self.fqdn_returns = []

    def select(self, ins, outs, excepts, timeout=0):
        self.select_calls.append((ins, outs, excepts, timeout))
        if self.select_fail > 0:
            self.select_fail -= 1
            raise select.error
        if not self.do_return:  # simplify code that doesn't need it
            self.do_return.append(True)
        doreturn = self.do_return.pop(0)
        if doreturn:
            return (ins, [], [])
        else:
            return ([], [], [])

    def getfqdn(self, name=""):
        self.fqdn_calls.append(name)
        return self.fqdn_returns.pop(0)


class path_mod:
    def __init__(self):
        self.isdir_calls = []
        self.isdir_returns = []
        self.getmtime_calls = []
        self.getmtime_returns = []
        self.join_calls = []

    def isdir(self, dirname):
        self.isdir_calls.append(dirname)
        return self.isdir_returns.pop(0)

    def join(self, *args):
        self.join_calls.append(args)
        return os.path.join(*args)

    def basename(self, pathname):
        self.join_calls.append(pathname)
        return os.path.basename(pathname)

    def getmtime(self, filename):
        self.getmtime_calls.append(filename)
        return self.getmtime_returns.pop(0)


class OSModuleJig:
    def __init__(self):
        self.isatty_calls = []
        self.isatty_returns = []
        self.path = path_mod()  # Need os.path

    def isatty(self, fd):
        self.isatty_calls.append(fd)
        return self.isatty_returns.pop(0)


class FcntlModuleJig:
    def __init__(self):
        self.ioctl_calls = []
        self.ioctl_returns = []

    def ioctl(self, fd, op, arg=0, mutate_flag=False):
        self.ioctl_calls.append((fd, op, arg, mutate_flag))
        return self.ioctl_returns.pop(0)


class ShutilModuleJig:
    def __init__(self):
        self.gts_calls = []
        self.gts_returns = []

    def get_terminal_size(self, default=(80, 24)):
        self.gts_calls.append(default)
        return self.gts_returns.pop(0)


class TimeModuleJig:
    def __init__(self):
        self.time_calls = 0
        self.time_returns = []

    def time(self):
        self.time_calls += 1
        return self.time_returns.pop(0)

    def strftime(self, f, t=None):
        if t is None:
            return time.strftime(f)
        else:
            return time.strftime(f, t)

    def gmtime(self, *args, **kwargs):
        return time.gmtime(*args, **kwargs)


class GzipModuleJig:
    def __init__(self):
        self.open_calls = []
        self.files_returned = []

    def open(self, filename, filemode):
        self.open_calls.append((filename, filemode))
        return self.files_returned.pop(0)


class GlobModuleJig:
    def __init__(self):
        self.glob_calls = []
        self.glob_returns = []

    def glob(self, pathname):
        self.glob_calls.append(pathname)
        ret = self.glob_returns.pop(0)
        return ret
